import socket
import sys


def bind_tcp_socket(host: str, port: int) -> socket.socket:
    family = socket.AF_INET

    if host and ":" in host:  # pragma: full coverage
        # It's an IPv6 address.
        family = socket.AF_INET6

    sock = socket.socket(family=family)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    try:
        sock.bind((host, port))
    except OSError:  # pragma: full coverage
        sys.exit(1)

    sock.setblocking(False)
    sock.set_inheritable(True)

    return sock


def bind_udp_socket(host: str, port: int) -> socket.socket:
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    try:
        sock.bind((host, port))

    except OSError:
        sys.exit(1)

    sock.setblocking(False)
    sock.set_inheritable(True)

    return sock



from __future__ import annotations
from typing import Any, Dict

import msgspec

from .log_level import LogLevel


class Entry(msgspec.Struct, kw_only=True):
    message: str | None = None
    tags: set[str] = msgspec.field(
        default_factory=set,
    )
    level: LogLevel

    def to_template(
        self,
        template: str,
        context: Dict[str, Any] | None = None,
    ):
        kwargs: Dict[
            str,
            int | str | bool | float | LogLevel | list | dict | set | Any,
        ] = {field: getattr(self, field) for field in self.__struct_fields__}

        kwargs["level"] = kwargs["level"].value

        if context:
            kwargs.update(context)

        return template.format(**kwargs)



import asyncio
import os

from typing import TypeVar, Any

from mkfst.logging.models import LogLevelName
from .logger_stream import LoggerStream
from .retention_policy import (
    RetentionPolicy,
    RetentionPolicyConfig,
)


T = TypeVar("T")


class LoggerContext:
    def __init__(
        self,
        name: str | None = None,
        template: str | None = None,
        filename: str | None = None,
        directory: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        nested: bool = False,
        models: dict[
            type[T],
            dict[str, Any],
        ]
        | None = None,
        level: LogLevelName | None = None,
    ) -> None:
        self.name = name
        self.template = template
        self.filename = filename
        self.directory = directory
        self.retention_policy = retention_policy
        self.level = level

        self.stream = LoggerStream(
            name=name,
            template=template,
            filename=filename,
            directory=directory,
            retention_policy=retention_policy,
            models=models,
            level=level,
        )
        self.nested = nested

    async def __aenter__(self):
        await self.stream.initialize()

        if self.stream._cwd is None:
            loop = asyncio.get_event_loop()
            self.stream._cwd = await loop.run_in_executor(
                None,
                os.getcwd,
            )

        if self.filename:
            await self.stream.open_file(
                self.filename,
                directory=self.directory,
                is_default=True,
                retention_policy=self.retention_policy,
            )

        if self.retention_policy and self.filename is None:
            filename = "logs.json"
            directory = os.path.join(self.stream._cwd, "logs")
            logfile_path = os.path.join(directory, filename)

            policy = RetentionPolicy(self.retention_policy)
            policy.parse()

            self.stream._retention_policies[logfile_path] = policy

        return self.stream

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.nested is False:
            await self.stream.close(
                shutdown_subscribed=self.stream.has_active_subscriptions
            )



from __future__ import annotations
import asyncio
import datetime

from mkfst import (
    HTML,
    Group,
    Service,
    endpoint,
    Model,
)


class MetadataV2(Model, tag=True):
    accessed: datetime.datetime


class MetadataV1(Model, tag=True):
    created: datetime.datetime
    updated: datetime.datetime


class UserV1(Model):
    username: str
    password: str
    metadata: MetadataV1


class UserV2(Model):
    username: str
    password: str
    metadata: MetadataV2


class UsersApiV1(Group):
    @endpoint("/get")
    async def get_service(self) -> UserV1:
        return UserV1(
            username="johnnyj",
            password="Password12345",
            metadata=MetadataV1(
                created=datetime.datetime.now(),
                updated=datetime.datetime.now(),
            ),
        )


class UsersApiV2(Group):
    @endpoint("/get")
    async def get_service(self) -> UserV2:
        return UserV2(
            username="johnnyj",
            password="Password12345",
            metadata=MetadataV2(accessed=datetime.datetime.now()),
        )


class ApiV1(Group):
    pass


class ApiV2(Group):
    pass


class TestService(Service):
    @endpoint("/home")
    async def get_home(self) -> HTML:
        return HTML(
            content="""
        <!DOCTYPE html>
        <html>
            <head>
                <title>Home</title>
            </head>
            <body>
                <h1>Hello from home!</h1>
            </body>
        </html>
        """
        )

    @endpoint("/")
    async def get_base(self) -> str:
        return "Hello World"


async def run_server():
    server = TestService(
        "localhost",
        5019,
        log_level="error",
        groups=[
            ApiV1("/api/v1", groups=[UsersApiV1("/users")]),
            ApiV2("/api/v2", groups=[UsersApiV2("/users")]),
        ],
        workers=0,
    )

    await server.run()


if __name__ == "__main__":
    asyncio.run(run_server())



from .http import HTML as HTML
from .http import Body as Body
from .http import Cookies as Cookies
from .http import FileUpload as FileUpload
from .http import Headers as Headers
from .http import Model as Model
from .http import Parameters as Parameters
from .http import Query as Query



from typing import (
    Any,
)

from urllib.parse import urlunparse
from mkfst.logging import Logger, LogLevel
from mkfst.middleware.base import Middleware
from mkfst.middleware.base.response_context import ResponseContext
from mkfst.middleware.base.types import Handler, MiddlewareHandler
from mkfst.models.logging import Event

ENFORCE_DOMAIN_WILDCARD = "Domain wildcard patterns must be like '*.example.com'."


class TrustedHost(Middleware):
    def __init__(
        self,
        allowed_hosts: list[str] | None = None,
        www_redirect: bool = True,
    ):
        self._logger = Logger()
        super().__init__(
            self.__class__.__name__,
        )

        if allowed_hosts is None:
            allowed_hosts = ["*"]

        for pattern in allowed_hosts:
            assert "*" not in pattern[1:], ENFORCE_DOMAIN_WILDCARD
            if pattern.startswith("*") and pattern != "*":
                assert pattern.startswith("*."), ENFORCE_DOMAIN_WILDCARD

        self.allowed_hosts = list(allowed_hosts)
        self.allow_any = "*" in allowed_hosts
        self.www_redirect = www_redirect

    async def __run__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ):
        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            await ctx.log(
                Event(
                    level=LogLevel.DEBUG,
                    message=f"Request - {context.method} {context.path}:{context.ip_address} - Checking for HTTPS redirect",
                )
            )

            try:
                if self.allow_any or context.protocol not in (
                    "http",
                    "ws",
                    "https",
                    "wss",
                ):
                    context.status = 400
                    context.errors.append(Exception("Unsupported protocol"))

                    await ctx.log(
                        Event(
                            level=LogLevel.ERROR,
                            message=f"Request - {context.method} {context.path}:{context.ip_address} - Invalid protocol ",
                        )
                    )

                    return (
                        context,
                        response,
                    ), False

                host: str | None = context.request_headers.get(
                    "host",
                    context.request_headers.get("Host"),
                )

                if host is None:
                    context.status = 400
                    context.errors.append(Exception("Invalid host header"))

                    await ctx.log(
                        Event(
                            level=LogLevel.ERROR,
                            message=f"Request - {context.method} {context.path}:{context.ip_address} - Invalid host header ",
                        )
                    )

                    return (
                        context,
                        response,
                    ), False

                host, _ = host.split(":", maxsplit=1)

                is_valid_host = False
                found_www_redirect = False

                for pattern in self.allowed_hosts:
                    if host == pattern or (
                        pattern.startswith("*") and host.endswith(pattern[1:])
                    ):
                        is_valid_host = True
                        break
                    elif "www." + host == pattern:
                        found_www_redirect = True

                if is_valid_host:
                    await ctx.log(
                        Event(
                            level=LogLevel.INFO,
                            message=f"Request - {context.method} {context.path}:{context.ip_address} - {host} matched host validation scheme",
                        )
                    )

                    return (
                        context,
                        response,
                    ), True

                else:
                    if found_www_redirect and self.www_redirect:
                        url = context.to_request_url()
                        redirect_url = urlunparse(
                            url._replace(
                                netloc="www." + url.netloc,
                            )
                        )

                        context.status = 307
                        context.response_headers.update(
                            {
                                "location": redirect_url,
                            }
                        )

                        await ctx.log(
                            Event(
                                level=LogLevel.INFO,
                                message=f"Request - {context.method} {context.path}:{context.ip_address} - Enforcing www redirect for {host} ",
                            )
                        )

                        return (
                            context,
                            response,
                        ), False

                    context.status = 400
                    context.errors.append(Exception("Invalid host header"))

                    await ctx.log(
                        Event(
                            level=LogLevel.ERROR,
                            message=f"Request - {context.method} {context.path}:{context.ip_address} - Invalid host header ",
                        )
                    )

                    return (
                        context,
                        response,
                    ), False

            except Exception as e:
                context.errors.append(e)
                context.status = 500

                await ctx.log(
                    Event(
                        level=LogLevel.ERROR,
                        message=f"Request - {context.method} {context.path}:{context.ip_address} - Encountered error attempting trusted host check - {str(e)}",
                    )
                )

                return (
                    context,
                    response,
                ), False

    async def close(self):
        pass

    def abort(self):
        pass



from typing import Dict, Literal

from pydantic import BaseModel, StrictStr

from .parsed_operation_metadata import ParsedOperationMetadata

HTTPMethod = Literal[
   "GET", "HEAD", "OPTIONS", "POST", "PUT", "PATCH", "DELETE", "TRACE"
]

class ParsedEndpointMetadata(BaseModel):
    description: StrictStr | None = None
    summary: StrictStr | None = None
    operations: Dict[HTTPMethod, ParsedOperationMetadata] = {}

    


from typing import (
    Any,
    Dict,
    List,
    Literal,
)

FieldType = Dict[Literal["type"], str]

FieldMetadata = Dict[
    Literal["title", "type", "format", "description", "anyOf"] | str,
    str | int | Any | List[FieldType],
]

FieldsMetadata = Dict[str, FieldMetadata]
PropertyMetadata = Dict[Literal["properties", "required"], FieldsMetadata | List[str]]

PropertySchema = dict[Literal["$defs"], dict[str, PropertyMetadata]]

SchemaType = Literal[
    "integer",
    "string",
    "boolean",
    "number",
    "array",
    "object",
]



from .backends.memory import SimpleMemoryCache as SimpleMemoryCache
from .base import BaseCache as BaseCache
from .decorators import cached as cached
from .decorators import cached_stampede as cached_stampede
from .decorators import multi_cached as multi_cached



import asyncio
import re
from typing import TypeVar, Generic
from typing import Callable, Literal, Tuple, TYPE_CHECKING
from .flow_control import FlowControl
from .server_state import ServerState
from .utils import (
    get_local_addr,
    get_remote_addr,
    is_ssl,
)
from .receive_buffer import ReceiveBuffer

T = TypeVar("T")


class MercurySyncTCPServerProtocol(asyncio.Protocol, Generic[T]):
    def __init__(self, conn: T):
        super().__init__()
        self.transport: asyncio.Transport = None
        self.loop = asyncio.get_event_loop()
        self.conn = conn
        self.flow: FlowControl = None
        self.server_state = ServerState[MercurySyncTCPServerProtocol]()
        self.connections = self.server_state.connections

        self.on_con_lost = self.loop.create_future()
        self.server: tuple[str, int] | None = None
        self.client: tuple[str, int] | None = None
        self.scheme: Literal["http", "https"] | None = None
        self.timeout_keep_alive_task: asyncio.TimerHandle | None = None

        self._receive_buffer = ReceiveBuffer()
        self._receive_buffer_closed = False
        self._active_requests: dict[bytes, bytes] = {}
        self._next_data: asyncio.Future = asyncio.Future()

    @property
    def trailing_data(self) -> tuple[bytes, bool]:
        """Data that has been received, but not yet processed, represented as
        a tuple with two elements, where the first is a byte-string containing
        the unprocessed data itself, and the second is a bool that is True if
        the receive connection was closed.

        See :ref:`switching-protocols` for discussion of why you'd want this.
        """
        return (bytes(self._receive_buffer), self._receive_buffer_closed)

    def connection_made(self, transport: asyncio.Transport) -> str:
        self.connections.add(self)
        self.transport = transport
        self.flow = FlowControl(transport)
        self.server = get_local_addr(transport)
        self.client = get_remote_addr(transport)
        self.scheme = "https" if is_ssl(transport) else "http"

    def data_received(self, data: bytes):
        if data:
            if self._receive_buffer_closed:
                raise RuntimeError("received close, then received more data?")
            self._receive_buffer += data
        else:
            self._receive_buffer_closed = True

        if self.conn.waiting_for_data.is_set() is False or (self._next_data.done()):
            self.conn.read(
                self._receive_buffer,
                self.transport,
                self._next_data,
            )
        else:
            self._next_data.set_result(self._receive_buffer)
            self._next_data = asyncio.Future()

    def connection_lost(self, exc: Exception | None):
        self.connections.discard(self)

        if self.flow is not None:
            self.flow.resume_writing()

        if exc is None:
            self.transport.close()
            # self._unset_keepalive_if_required()

        self.on_con_lost.set_result(True)

    def _unset_keepalive_if_required(self) -> None:
        if self.timeout_keep_alive_task is not None:
            self.timeout_keep_alive_task.cancel()
            self.timeout_keep_alive_task = None

    def eof_received(self) -> None:
        pass

    def on_response_complete(self) -> None:
        self.server_state.total_requests += 1

        if self.transport.is_closing():
            return

        # Set a short Keep-Alive timeout.
        # self._unset_keepalive_if_required()

        # self.timeout_keep_alive_task = self.loop.call_later(
        #     self.timeout_keep_alive, self.timeout_keep_alive_handler
        # )

        # Unpause data reads if needed.
        self.flow.resume_reading()

    def shutdown(self) -> None:
        """
        Called by the server to commence a graceful shutdown.
        """
        pass

    def pause_writing(self) -> None:
        """
        Called by the transport when the write buffer exceeds the high water mark.
        """
        self.flow.pause_writing()  # pragma: full coverage

    def resume_writing(self) -> None:
        """
        Called by the transport when the write buffer drops below the low water mark.
        """
        self.flow.resume_writing()  # pragma: full coverage



from __future__ import annotations
import time
from typing import Any, Optional
import msgspec
from .run_status import RunStatus
from .task_type import TaskType


class TaskRun(msgspec.Struct):
    run_id: int
    task_name: str
    status: RunStatus
    error: Optional[str] = None
    trace: Optional[str] = None
    start: int | float = time.monotonic()
    end: Optional[int | float] = None
    elapsed: int | float = 0
    result: Optional[Any] = None
    task_type: TaskType = TaskType.CALLABLE

    def complete(self):
        return self.status in [
            RunStatus.COMPLETE,
            RunStatus.CANCELLED,
            RunStatus.FAILED,
        ]



import asyncio
import socket
import ssl
from collections import defaultdict, deque
from typing import (
    Any,
    Coroutine,
    Deque,
    Dict,
    Optional,
    Tuple,
    Union,
    Callable,
)

import msgspec
import zstandard

from mkfst.connection.base.connection_type import ConnectionType
from mkfst.connection.tcp.protocols import (
    MercurySyncTCPServerProtocol,
)
from mkfst.encryption import AESGCMFernet
from mkfst.env import Env
from mkfst.env.time_parser import TimeParser
from mkfst.snowflake.snowflake_generator import SnowflakeGenerator


class MercurySyncTCPConnection:
    def __init__(self, host: str, port: int, instance_id: int, env: Env) -> None:
        self.id_generator = SnowflakeGenerator(instance_id)
        self.env = env

        self.host = host
        self.port = port

        self.events: Dict[str, Coroutine] = {}

        self.queue: Dict[str, Deque[Tuple[str, int, float, Any]]] = defaultdict(deque)
        self.connected = False
        self._running = False

        self._client_transports: Dict[str, asyncio.Transport] = {}
        self._server: asyncio.Server = None
        self._upgrade_server: asyncio.Server | None = None
        self._loop: Union[asyncio.AbstractEventLoop, None] = None
        self._waiters: Dict[str, Deque[asyncio.Future]] = defaultdict(deque)
        self._pending_responses: Deque[asyncio.Task] = deque()
        self._last_call: Deque[str] = deque()

        self._sent_values = deque()
        self.server_socket: socket.socket | None = None
        self.upgrade_socket: socket.socket | None = None
        self._stream = False

        self._client_key_path: Union[str, None] = None
        self._client_cert_path: Union[str, None] = None

        self._server_key_path: Union[str, None] = None
        self._server_cert_path: Union[str, None] = None

        self._client_ssl_context: Union[ssl.SSLContext, None] = None
        self._server_ssl_context: Union[ssl.SSLContext, None] = None

        self._encryptor = AESGCMFernet(env)
        self._semaphore: Union[asyncio.Semaphore, None] = None
        self._compressor: Union[zstandard.ZstdCompressor, None] = None
        self._decompressor: Union[zstandard.ZstdDecompressor, None] = None
        self._cleanup_task: Union[asyncio.Task, None] = None
        self._sleep_task: Union[asyncio.Task, None] = None
        self._cleanup_interval = TimeParser(env.MERCURY_SYNC_CLEANUP_INTERVAL).time

        self._request_timeout = TimeParser(env.MERCURY_SYNC_REQUEST_TIMEOUT).time

        self._max_concurrency = env.MERCURY_SYNC_MAX_CONCURRENCY
        self._tcp_connect_retries = env.MERCURY_SYNC_TCP_CONNECT_RETRIES
        self._verify_cert = env.MERCURY_SYNC_VERIFY_SSL_CERT

        self.connection_type = ConnectionType.TCP

    def from_env(self, env: Env):
        self._max_concurrency = env.MERCURY_SYNC_MAX_CONCURRENCY
        self._tcp_connect_retries = env.MERCURY_SYNC_TCP_CONNECT_RETRIES
        self._verify_cert = env.MERCURY_SYNC_VERIFY_SSL_CERT

    async def connect_async(
        self,
        cert_path: Optional[str] = None,
        key_path: Optional[str] = None,
        worker_socket: Optional[socket.socket] = None,
        upgrade_socket: Optional[socket.socket] = None,
        worker_server: Optional[asyncio.Server] = None,
    ):
        try:
            self._loop = asyncio.get_event_loop()

        except Exception:
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)

        self._running = True
        self._semaphore = asyncio.Semaphore(self._max_concurrency)

        self._compressor = zstandard.ZstdCompressor()
        self._decompressor = zstandard.ZstdDecompressor()

        if cert_path and key_path:
            self._server_ssl_context = self._create_server_ssl_context(
                cert_path=cert_path,
                key_path=key_path,
            )

        if self.connected is False and worker_socket is None:
            self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            try:
                self.server_socket.bind((self.host, self.port))

            except Exception:
                pass

            self.server_socket.setblocking(False)

        elif self.connected is False and worker_socket:
            self.server_socket = worker_socket
            host, port = worker_socket.getsockname()

            self.host = host
            self.port = port

        elif self.connected is False and worker_server:
            self._server = worker_server

            server_socket, _ = worker_server.sockets
            host, port = server_socket.getsockname()
            self.host = host
            self.port = port

            self.connected = True
            self._cleanup_task = self._loop.create_task(self._cleanup())

        if self.connected is False and upgrade_socket:
            self.upgrade_socket = upgrade_socket
            host, port = upgrade_socket.getsockname()

            self.host = host
            self.port = port

        if self.connected is False:
            server = await self._loop.create_server(
                lambda: MercurySyncTCPServerProtocol(self),
                sock=self.server_socket,
                ssl=self._server_ssl_context if self.upgrade_socket is None else None,
            )

            if self.upgrade_socket:
                upgrade_server = await self._loop.create_server(
                    lambda: MercurySyncTCPServerProtocol(self),
                    sock=self.upgrade_socket,
                    ssl=self._server_ssl_context,
                )

                self._upgrade_server = upgrade_server

            self._server = server
            self.connected = True

            self._cleanup_task = self._loop.create_task(self._cleanup())

    def _create_server_ssl_context(
        self,
        cert_path: Optional[str] = None,
        key_path: Optional[str] = None,
    ) -> ssl.SSLContext:
        if self._server_cert_path is None:
            self._server_cert_path = cert_path

        if self._server_key_path is None:
            self._server_key_path = key_path

        ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        ssl_ctx.options |= ssl.OP_NO_TLSv1
        ssl_ctx.options |= ssl.OP_NO_TLSv1_1
        ssl_ctx.options |= ssl.OP_SINGLE_DH_USE
        ssl_ctx.options |= ssl.OP_SINGLE_ECDH_USE
        ssl_ctx.load_cert_chain(cert_path, keyfile=key_path)
        ssl_ctx.load_verify_locations(cafile=cert_path)
        ssl_ctx.check_hostname = False

        match self._verify_cert:
            case "REQUIRED":
                ssl_ctx.verify_mode = ssl.VerifyMode.CERT_REQUIRED

            case "OPTIONAL":
                ssl_ctx.verify_mode = ssl.VerifyMode.CERT_OPTIONAL

            case _:
                ssl_ctx.verify_mode = ssl.VerifyMode.CERT_NONE

        ssl_ctx.set_ciphers("ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384")

        return ssl_ctx

    def read(self, data: bytes, transport: asyncio.Transport):
        pass

    async def _cleanup(self):
        while self._running:
            self._sleep_task = asyncio.create_task(
                asyncio.sleep(self._cleanup_interval)
            )

            try:
                await self._sleep_task

            except (Exception, asyncio.CancelledError, KeyboardInterrupt):
                pass

            for pending in list(self._pending_responses):
                if pending.done() or pending.cancelled():
                    try:
                        await pending

                    except (Exception, socket.error):
                        pass
                    self._pending_responses.pop()

    async def close(self) -> None:
        self._stream = False
        self._running = False

        for client in self._client_transports.values():
            client.abort()

        if self._cleanup_task:
            self._cleanup_task.cancel()
            if self._cleanup_task.cancelled() is False:
                try:
                    self._sleep_task.cancel()
                    if not self._sleep_task.cancelled():
                        await self._sleep_task

                except (Exception, socket.error):
                    pass

                try:
                    await self._cleanup_task

                except Exception:
                    pass

    def abort(self) -> None:
        self._stream = False
        self._running = False

        if self._cleanup_task:
            try:
                self._sleep_task.cancel()

            except (
                asyncio.CancelledError,
                asyncio.InvalidStateError,
                asyncio.TimeoutError,
                Exception,
                socket.error,
            ):
                pass

            try:
                self._cleanup_task.cancel()

            except (
                asyncio.CancelledError,
                asyncio.InvalidStateError,
                asyncio.TimeoutError,
                Exception,
                socket.error,
            ):
                pass



from base64 import b64encode
from typing import (
    Any,
    Callable,
    Dict,
    Union,
)

import zstandard

from mkfst.logging import Logger, LogLevel
from mkfst.middleware.base import Middleware, MiddlewareType
from mkfst.middleware.base.response_context import ResponseContext
from mkfst.middleware.base.types import Handler, MiddlewareHandler, MiddlewareResult
from mkfst.models.http.parse_response import parse_response
from mkfst.models.logging import Event


class BidirectionalZStandardCompressor(Middleware):
    def __init__(
        self,
        compression_level: int = 9,
        serializers: Dict[
            str, Callable[..., Union[str, None]]
        ] = {},
    ) -> None:
        super().__init__(
            self.__class__.__name__, 
            middleware_type=MiddlewareType.BIDIRECTIONAL,
            response_headers={
                "x-compression-encoding": "zstd"
            }
        )

        self.compression_level = compression_level
        self.serializers = serializers
        self._compressor = zstandard.ZstdCompressor()
        self._logger = Logger()

    async def __pre__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ) -> MiddlewareResult:

        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            
            await ctx.log(Event(
                level=LogLevel.DEBUG,
                message=f'Request - {context.method} {context.path}:{context.ip_address} - Compressing request via ZStd with level {self.compression_level} compression'
            ))
            
            data = context.get_bytes_arg()
            compressed_data = b''
            try:
                if data != b"":
                    compressed_data = self._compressor.compress(
                        data, 
                        compresslevel=self.compression_level,
                    )

                    context.update_request_data(compressed_data)
                    context.update_request_headers({
                        "x-compression-encoding": "zstd"
                    })

                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Compressed request via ZStd with level {self.compression_level} compression'
                    ))

                else:
                    await ctx.log(Event(
                        level=LogLevel.WARN,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - No request to compress'
                    ))

                return (
                    context,
                    response
                ), True

            except Exception as e:
                context.errors.append(e)
                context.compressor = 'zstd'
                context.compression_level = self.compression_level
                context.status = 500

                await ctx.log(Event(
                    level=LogLevel.ERROR,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Encountered error ZStd compressing request - {str(e)}'
                ))

                return (
                    context,
                    response
                ), False

    async def __post__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ) -> MiddlewareResult:
        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            try:
                if response is None:
                    await ctx.log(Event(
                        level=LogLevel.WARN,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - No response to compress'
                    ))

                    return (
                        context,
                        response
                    ), True

                elif isinstance(response, str):
                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Compressing string response via ZStd with level {self.compression_level} compression'
                    ))

                    compressed_data = self._compressor.compress(
                        response.encode(),
                    )

                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Compressed response via ZStd with level {self.compression_level} compression'
                    ))

                    return (
                        context,
                        b64encode(compressed_data).decode()
                    ), True

                else:
                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Compressing response via ZStd with level {self.compression_level} compression'
                    ))

                    serialized = parse_response(
                        response,
                        context.parser,
                    )

                    compressed_data = self._compressor.compress(serialized)
                    context.response_headers["x-compression-encoding"] = "zstd"

                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Compressed response via ZStd with level {self.compression_level} compression'
                    ))

                    return (
                        context,
                        b64encode(compressed_data).decode()
                    ), True

            except Exception as e:
                context.compressor = 'zstd'
                context.compression_level = self.compression_level
                context.errors.append(e)
                context.status = 500

                await ctx.log(Event(
                    level=LogLevel.ERROR,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Encountered error ZStd compressing request - {str(e)}'
                ))

                return (
                    context,
                    response,
                ), False

    async def close(self):
        await self._logger.close()

    def abort(self):
        self._logger.abort()


from __future__ import annotations
import time
from typing import Any, Optional

import orjson
import msgspec
from .run_status import RunStatus


class TaskRun(msgspec.Struct, kw_only=True):
    run_id: int
    status: RunStatus
    error: Optional[str] = None
    trace: Optional[str] = None
    start: int | float = time.monotonic()
    end: Optional[int | float] = None
    elapsed: int | float = 0
    result: Optional[Any] = None

    def to_json(self):
        return orjson.dumps(
            {
                "run_id": self.run_id,
                "status": self.status.value,
                "error": self.error,
                "start": self.start,
                "end": self.end,
                "elapsed": self.elapsed,
            }
        )

    def to_data(self):
        return {
            "run_id": self.run_id,
            "status": self.status.value,
            "error": self.error,
            "start": self.start,
            "end": self.end,
            "elapsed": self.elapsed,
        }



import re

class FileSizeParser:

    def __init__(self) -> None:      
        self._units = {
            "B": 1, 
            "KB": 2**10, 
            "MB": 2**20, 
            "GB": 2**30, 
            "TB": 2**40
        }


    def parse(self, file_size: str):
        file_size = file_size.upper()

        if not re.match(r'([KMGT]?B)', file_size):
            file_size = re.sub(
                r'([KMGT]?B)', 
                r' \1', 
                file_size
            )

        number, unit = [string.strip() for string in file_size.split()]

        return int(
            float(number) * self._units[unit]
        )


from pydantic import BaseModel, RootModel, StrictStr, conlist


class InternalError(BaseModel):
    error: StrictStr


class InternalErrorSet(RootModel):
    root: conlist(InternalError, min_length=1)


class BadRequestError(BaseModel):
    error: str


class BadRequestErrorSet(RootModel):
    root: conlist(BadRequestError, min_length=1)



from typing import (
    Any,
    TypeVar,
    Generic,
)
from mkfst.logging import Logger, LogLevel
from mkfst.middleware.base import Middleware
from mkfst.middleware.base.response_context import ResponseContext
from mkfst.middleware.base.types import Handler, MiddlewareHandler
from mkfst.models.logging import Event


T = TypeVar("T")


class Authentication(Middleware, Generic[T]):
    def __init__(self, authenticator: T):
        self._logger = Logger()
        super().__init__(
            self.__class__.__name__,
            methods=[
                "DELETE",
                "GET",
                "HEAD",
                "OPTIONS",
                "PATCH",
                "POST",
                "PUT",
                "TRACE",
            ],
        )

        self._authenticator = authenticator

    async def __run__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ):
        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            await ctx.log(
                Event(
                    level=LogLevel.DEBUG,
                    message=f"Request - {context.method} {context.path}:{context.ip_address} - Executing authentication",
                )
            )

            try:
                pass

            except Exception as e:
                context.errors.append(e)
                context.status = 500

                await ctx.log(
                    Event(
                        level=LogLevel.ERROR,
                        message=f"Request - {context.method} {context.path}:{context.ip_address} - Encountered error attempting authenticate - {str(e)}",
                    )
                )

                return (
                    context,
                    response,
                ), False



from __future__ import annotations
import msgspec
import orjson
from typing import TypeVar


T = TypeVar("T", bound=dict)


class Model(msgspec.Struct):
    @classmethod
    def model_json_schema(cls, ref_template: str | None = None):
        if ref_template:
            components: tuple[dict[str, dict]] = msgspec.json.schema_components(
                [cls],
                ref_template=ref_template.format(model=cls.__name__),
            )

            schema_properties = {}

            for item in components:
                if isinstance(item, dict):
                    schema_properties.update(
                        item.get(
                            "$defs",
                            {},
                        ).get(cls.__name__, {})
                    )

            return schema_properties

        else:
            component: dict[str, dict] = msgspec.json.schema(cls)
            return component.get(
                "$defs",
                {},
            ).get(cls.__name__, {})

    @classmethod
    def defaults(cls):
        return {field.name: field.default for field in msgspec.structs.fields(cls)}

    @classmethod
    def model_fields(cls):
        return {field.name: field for field in msgspec.structs.fields(cls)}

    @classmethod
    def model_spec(cls):
        return {field.name: field.type for field in msgspec.structs.fields(cls)}

    def model_dump(self, exclude_none: bool = False):
        if exclude_none:
            return {
                key: value.model_dump() if isinstance(value, Model) else value
                for key, value in msgspec.structs.asdict(self).items()
                if value is not None
            }

        return {
            key: value.model_dump() if isinstance(value, Model) else value
            for key, value in msgspec.structs.asdict(self).items()
        }

    def model_dump_json(self, exclude_none: bool = False):
        if exclude_none:
            return orjson.dumps(
                {
                    key: value.model_dump() if isinstance(value, Model) else value
                    for key, value in msgspec.structs.asdict(self).items()
                    if value is not None
                }
            )

        return orjson.dumps(
            {
                key: value.model_dump() if isinstance(value, Model) else value
                for key, value in msgspec.structs.asdict(self).items()
            }
        )

    def model_copy(self, update: T):
        model_dict = msgspec.structs.asdict(self)
        model_dict.update(update)

        return type(self)(**model_dict)



import asyncio
from typing import List

from mkfst.logging.models import Log

from .consumer_status import ConsumerStatus
from .log_consumer import LogConsumer
from .provider_status import ProviderStatus


class LogProvider:
    def __init__(self) -> None:
        self._close_waiter: asyncio.Future | None = None
        self.closing: bool = False
        self._consumers: List[LogConsumer] = []
        self.status = ProviderStatus.READY

    @property
    def subscriptions_count(self):
        return len(self._consumers)

    def subscribe(self, consumer: LogConsumer):
        if self.status == ProviderStatus.READY:
            self.status = ProviderStatus.RUNNING

        if self.status == ProviderStatus.RUNNING:
            self._consumers.append(consumer)

    async def put(self, log: Log):
        if self.status == ProviderStatus.RUNNING:
            await asyncio.gather(
                *[
                    consumer.put(log)
                    for consumer in self._consumers
                    if consumer.status
                    in [
                        ConsumerStatus.READY,
                        ConsumerStatus.RUNNING,
                    ]
                ]
            )

        await asyncio.sleep(0)

    async def signal_shutdown(self):
        self.status = ProviderStatus.CLOSING

        for consumer in self._consumers:
            consumer.stop()

            if consumer.pending:
                await consumer.wait_for_pending()

        self.status = ProviderStatus.CLOSED



from typing import Any, Dict, List

from .endpoint_parser import EndpointParser
from .models import (
    ServerVariable,
)
from .parsed_api_metadata import ParsedAPIMetadata
from .parsed_api_server import ParsedAPIServer
from .parsed_endpoint import ParsedEndpoint
from .parsed_tag import ParsedTag


def remove_none(obj):
    if isinstance(obj, (list, tuple, set)):
        return type(obj)(remove_none(x) for x in obj if x is not None)
    elif isinstance(obj, dict):
        return type(obj)(
            (remove_none(k), remove_none(v))
            for k, v in obj.items()
            if k is not None and v is not None
        )
    else:
        return obj


def create_api_definition(
    api_metadata: ParsedAPIMetadata,
    api_config: ParsedAPIServer,
    endpoints: Dict[str, ParsedEndpoint],
    tags: List[ParsedTag],
):
    endpoint_parsers = [
        EndpointParser(**parsed.to_dict(), default_tags=tags)
        for parsed in endpoints.values()
    ]

    server_variables: Dict[str, ServerVariable] | None = None
    if api_config.server_variables:
        server_variables = {
            name: {
                "enum": variable.options,
                "default": variable.default,
                "description": variable.description,
            }
            for name, variable in api_config.server_variables.items()
        }
    schema_components: Dict[str, Any] = {}
    paths: Dict[str, Any] = {}
    for parser in endpoint_parsers:
        paths[parser.cleaned_path] = parser.parse()
        schema_components.update(parser.request_components)
        schema_components.update(parser.response_components)

    schema = {
        "openapi": "3.1.2",
        "tags": [tag.parse() for tag in tags],
        "info": {
            "title": api_metadata.title,
            "summary": api_metadata.summary,
            "description": api_metadata.description,
            "termsOfService": api_metadata.terms_of_service,
            "contact": {
                "name": api_metadata.owner,
                "url": api_metadata.owner_url.unicode_string()
                if api_metadata.owner_url
                else None,
                "email": api_metadata.owner_email,
            },
            "license": {
                "name": api_metadata.license,
                "identifier": api_metadata.license_identifier,
                "url": api_metadata.license_url.unicode_string()
                if api_metadata.license_url
                else None,
            }
            if api_metadata.license
            else None,
            "version": api_metadata.version,
        },
        "servers": [
            {
                "url": api_config.server_url.unicode_string(),
                "description": api_config.server_description,
                "variables": server_variables,
            }
        ],
        "components": {"schemas": schema_components},
        "paths": {parser.cleaned_path: parser.parse() for parser in endpoint_parsers},
    }

    return remove_none(schema)



import functools
from collections import defaultdict
from typing import (
    Dict,
    List,
    Literal,
    Optional,
    TypeVar,
    get_args,
    get_type_hints,
)

from mkfst.models.http import Limit, Model

from .get_content_type import get_content_type

T = TypeVar("T")


def endpoint(
    path: Optional[str] = "/",
    methods: List[
        Literal[
            "GET",
            "HEAD",
            "OPTIONS",
            "POST",
            "PUT",
            "PATCH",
            "DELETE",
            "TRACE",
        ],
    ] = ["GET"],
    responses: Optional[Dict[int, Model]] = None,
    response_headers: Optional[Dict[str, str]] = None,
    limit: Optional[Limit] = None,
    summary: str | None = None,
    tags: List[str] | None = None,
    method_metadata: Dict[
        Literal[
            "GET",
            "HEAD",
            "OPTIONS",
            "POST",
            "PUT",
            "PATCH",
            "DELETE",
            "TRACE",
        ],
        Dict[
            Literal[
                "name",
                "tags",
                "description",
                "summary",
                "docs_url",
                "docs_description",
                "depreciated",
            ],
            str | List[str] | bool,
        ],
    ]
    | None = None,
    additional_docs: Dict[Literal["doc_description", "docs_url"], str] | None = None,
    depreciated: bool = False,
):
    def wraps(func):
        return_type = get_type_hints(func).get("return")
        return_args = get_args(return_type)

        if isinstance(return_args, tuple) and len(return_args) > 0:
            return_type = return_args[0]

        headers = response_headers or {}

        lowered_headers: Dict[str, str] = {}
        lowered_headers = {key.lower(): value.lower() for key, value in headers.items()}

        if lowered_headers.get("content-type") is None:
            headers.update({"content-type": get_content_type(return_type)})

        func.as_endpoint = True
        func.path = path
        func.methods = methods

        func.response_headers = headers
        func.responses = responses
        func.limit = limit
        func.summary = summary
        func.tags = tags
        func.method_metadata = method_metadata or defaultdict(dict)
        func.additional_docs = (additional_docs,)
        func.depreciated = depreciated

        @functools.wraps(func)
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return decorator

    return wraps



from .log_consumer import LogConsumer as LogConsumer
from .log_provider import LogProvider as LogProvider
from .consumer_status import ConsumerStatus as ConsumerStatus
from .provider_status import ProviderStatus as ProviderStatus


from .env import Env as Env
from .hooks import endpoint as endpoint
from .hooks import task as task
from .models.http import HTML as HTML
from .models.http import Body as Body
from .models.http import Cookies as Cookies
from .models.http import FileUpload as FileUpload
from .models.http import Headers as Headers
from .models.http import (
    Limit as Limit,
)
from .models.http import Model as Model
from .models.http import Parameters as Parameters
from .models.http import Query as Query
from .service import Group as Group
from .service import Service as Service



import asyncio


def get_remote_addr(transport: asyncio.Transport) -> tuple[str, int] | None:
    socket_info = transport.get_extra_info("socket")
    if socket_info is not None:
        try:
            info = socket_info.getpeername()
            return (str(info[0]), int(info[1])) if isinstance(info, tuple) else None
        except OSError:  # pragma: no cover
            # This case appears to inconsistently occur with uvloop
            # bound to a unix domain socket.
            return None

    info = transport.get_extra_info("peername")
    if info is not None and isinstance(info, (list, tuple)) and len(info) == 2:
        return (str(info[0]), int(info[1]))
    return None


def get_local_addr(transport: asyncio.Transport) -> tuple[str, int] | None:
    socket_info = transport.get_extra_info("socket")
    if socket_info is not None:
        info = socket_info.getsockname()

        return (str(info[0]), int(info[1])) if isinstance(info, tuple) else None
    info = transport.get_extra_info("sockname")
    if info is not None and isinstance(info, (list, tuple)) and len(info) == 2:
        return (str(info[0]), int(info[1]))
    return None


def is_ssl(transport: asyncio.Transport) -> bool:
    return bool(transport.get_extra_info("sslcontext"))




from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
)

import msgspec

from .base_wrapper import BaseWrapper
from .response_context import ResponseContext
from .types import Handler, MiddlewareHandler, MiddlewareType

T = TypeVar("T")


class CallWrapper(BaseWrapper):
    def __init__(
        self,
        name: str,
        handler: Handler,
        middleware_type: MiddlewareType = MiddlewareType.CALL,
        methods: Optional[
            List[
                Literal[
                    "GET", "HEAD", "OPTIONS", "POST", "PUT", "PATCH", "DELETE", "TRACE"
                ]
            ]
        ] = None,
        responses: Optional[Dict[int, msgspec.Struct]] = None,
        serializers: Optional[Dict[int, Callable[..., str]]] = None,
        response_headers: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__()

        self.name = name
        self.path = handler.path
        self.methods: List[
            Literal["GET", "HEAD", "OPTIONS", "POST", "PUT", "PATCH", "DELETE", "TRACE"]
        ] = handler.methods

        if methods:
            self.methods.extend(methods)

        self.response_headers: Union[Dict[str, str], None] = handler.response_headers

        if self.response_headers and response_headers:
            self.response_headers.update(response_headers)

        elif response_headers:
            self.response_headers = response_headers

        self.responses = responses
        self.serializers = serializers
        self.limit = handler.limit

        self.handler = handler
        self.wraps = isinstance(handler, BaseWrapper)

        self.run: Optional[MiddlewareHandler] = None

        self.middleware_type = middleware_type

    async def __call__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
    ):
        (context, response), _ = await self.run(
            context=context,
            response=response,
            handler=self.handler,
        )

        return context, response



import asyncio
import functools
from typing import Literal
from .model import Model


class FileUpload(Model):
    data: bytes
    content_type: str | None = None
    encoding: str | None = None

    async def upload(
        self,
        path: str,
        read_type: Literal["string", "binary"] = "string",
    ):
        upload_file = await asyncio.to_thread(
            functools.partial(self._upload, path, read_type)
        )

        return FileUpload(data=upload_file)

    def _upload(
        self,
        path: str,
        read_type: Literal["string", "binary"] = "string",
    ):
        read_mode = "r"
        if read_type == "binary":
            read_mode = "rb"

        with open(path, read_mode) as upload_file:
            return upload_file.read()



from __future__ import annotations
from urllib.parse import urlparse, ParseResult
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Tuple,
    Type,
)

from mkfst.connection.tcp.fabricator import Fabricator
from mkfst.models.http.request_models import Cookies, Headers


class ResponseContext:
    __slots__ = (
        "path",
        "method",
        "params",
        "query",
        "request_headers",
        "cookies",
        "body",
        "response_headers",
        "status",
        "errors",
        "args",
        "kwargs",
        "_data",
        "fabricator",
        "parser",
        "compressor",
        "compression_level",
        "ip_address",
        "request_addr",
        "protocol",
        "upgrade_port",
    )

    def __init__(
        self,
        path: str,
        method: str,
        headers: dict[str, str],
        params: Dict[str, str] | None,
        query: str | None,
        data: List[bytes],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        fabricator: Fabricator,
        parser: Type[Any],
        ip_address: str,
        protocol: Literal["https", "http", "wss", "ws"],
        request_addr: tuple[str, int],
        upgrade_port: int,
    ) -> None:
        self.ip_address = ip_address
        self.protocol = protocol
        self.path = path
        self.method = method
        self.params: Dict[str, Any] = params or {}
        self.query: str = query or ""

        self.request_headers: Dict[str, Any] = headers
        self.cookies: Dict[str, Any] = {}
        self.body: Any | None = None

        self.response_headers: Dict[str, Any] = {}
        self.status: int | None = None
        self.parser = parser
        self.errors: List[Exception] = []

        self.args = args
        self.kwargs = kwargs
        self._data = data
        self.fabricator = fabricator
        self.compressor: Literal["gzip", "zstd"] | None = None
        self.compression_level: int | None = None
        self.request_addr = request_addr
        self.upgrade_port = upgrade_port

    def update(self, context: ResponseContext):
        self.request_headers.update(context.request_headers)
        self.response_headers.update(context.response_headers)
        self.params.update(context.params)
        self.cookies.update(context.cookies)

        if context.status:
            self.status = context.status

    def to_request_url(self) -> ParseResult:
        host: str = self.request_headers.get("host")
        port: int | None = None
        if host is None:
            host, port = self.request_addr

        scheme = self.protocol

        if port:
            url = f"{scheme}://{host}:{port}"

        else:
            url = f"{scheme}://{host}"

        if self.path:
            url += self.path

        return urlparse(url)

    def get_bytes_arg(
        self,
    ) -> Tuple[bytes, int, None] | Tuple[bytes, None, int] | Tuple[bytes, None, None]:
        for arg in self.args:
            if isinstance(arg, bytes):
                return arg

        for value in self.kwargs.values():
            if isinstance(value, bytes):
                return value

        return b""

    def get_headers(self):
        if param_key := self.fabricator.param_keys.get("headers"):
            is_position = isinstance(param_key, int)
            headers: Headers = (
                self.args[param_key] if is_position else self.kwargs[param_key]
            )

            return headers

        return None

    def get_headers_and_cookies(self):
        headers: Dict[str, Any] = {}
        cookies: Dict[str, Any] = {}

        if param_key := self.fabricator.param_keys.get("headers"):
            headers: Headers = (
                self.args[param_key]
                if isinstance(param_key, int)
                else self.kwargs[param_key]
            )

        elif param_key := self.fabricator.param_keys.get("cookies"):
            cookies: Cookies = (
                self.args[param_key]
                if isinstance(param_key, int)
                else self.kwargs[param_key]
            )

        headers_content: Dict[str, str] = {}
        cookies_content: Dict[str, str] = {}
        if headers:
            headers_content = headers.model_dump()

        if headers and cookies is None:
            cookies_content = Cookies.make_raw(headers)

        elif cookies:
            cookies_content = cookies.model_dump()

        return (
            headers_content,
            cookies_content,
        )

    def update_request_headers(self, new_headers: Dict[str, Any]):
        self.request_headers.update(new_headers)

        headers = self.get_headers()

        updated_headers: Headers | None = None

        if headers:
            headers_dict = headers.model_dump()

            headers_dict.update(self.request_headers)

            updated_headers = headers.model_copy(headers_dict)

        param_key = self.fabricator.param_keys.get("headers")
        is_position = isinstance(param_key, int)

        if updated_headers and param_key and is_position:
            self.args[param_key] = updated_headers

        elif updated_headers and param_key:
            self.kwargs[param_key] = updated_headers

    def update_request_data(self, data: Any):
        body_key = self.fabricator.param_keys.get("body")
        is_position = isinstance(body_key, int)

        if body_key and is_position:
            self.args[body_key] = data

        elif body_key:
            self.kwargs[body_key] = data



from re import Pattern
from typing import Any


def validate(
    regex: Pattern[bytes], data: bytes, msg: str = "malformed data", *format_args: Any
) -> dict[str, bytes]:
    match = regex.fullmatch(data)

    if not match:
        if format_args:
            msg = msg.format(*format_args)

        raise Exception(msg)

    return match.groupdict()



import asyncio
import math
import random
from typing import (
    Any,
    Optional,
    Union,
)

from mkfst.env import Env, load_env
from mkfst.env.time_parser import TimeParser
from mkfst.logging import Logger, LogLevel
from mkfst.middleware.base import Middleware, MiddlewareType
from mkfst.middleware.base.response_context import ResponseContext
from mkfst.middleware.base.types import Handler, MiddlewareHandler, MiddlewareResult
from mkfst.models.logging import Event
from mkfst.rate_limiting.limiters import SlidingWindowLimiter

from .circuit_breaker_state import CircuitBreakerState


class CircuitBreaker(Middleware):
    def __init__(
        self,
        failure_threshold: Optional[float] = None,
        failure_window: Optional[str] = None,
        handler_timeout: Optional[str] = None,
        rejection_sensitivity: Optional[float] = None,
    ) -> None:
        super().__init__(
            self.__class__.__name__, 
            middleware_type=MiddlewareType.CALL,
            response_headers={
                "x-mercury-sync-overload": True
            }
        )

        env = load_env(Env)

        self._logger = Logger()

        if failure_threshold is None:
            failure_threshold = env.MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_THRESHOLD

        if failure_window is None:
            failure_window = env.MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_WINDOW

        if handler_timeout is None:
            handler_timeout = env.MERCURY_SYNC_HTTP_HANDLER_TIMEOUT

        if rejection_sensitivity is None:
            rejection_sensitivity = (
                env.MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_REJECTION_SENSITIVITY
            )

        self.failure_threshold = failure_threshold
        self.rejection_sensitivity = rejection_sensitivity

        self.failure_window = TimeParser(failure_window).time
        self.handler_timeout = TimeParser(handler_timeout).time
        self._limiter_failure_window = failure_window

        self.overload = 0
        self.failed = 0
        self.succeeded = 0
        self.total_completed = 0

        self._rate_per_sec = 0
        self._rate_per_sec_succeeded = 0
        self._rate_per_sec_failed = 0

        self._previous_count = 0
        self._previous_count_succeeded = 0
        self._previous_count_failed = 0

        self.wraps: bool = False

        self._loop: Union[asyncio.AbstractEventLoop, None] = None
        self._current_time: Union[float, None] = None
        self._breaker_state = CircuitBreakerState.CLOSED

        self._limiter: Union[SlidingWindowLimiter, None] = None

        self._closed_window_start: Union[float, None] = None
        self._closed_elapsed = 0

        self._half_open_window_start: Union[float, None] = None
        self._half_open_elapsed = 0

    def trip_breaker(self) -> bool:
        failed_rate_threshold = max(self._rate_per_sec * self.failure_threshold, 1)

        return int(self._rate_per_sec_failed) > int(failed_rate_threshold)

    def reject_request(self) -> bool:
        if (self._loop.time() - self._current_time) > self.failure_window:
            self._current_time = (
                math.floor(self._loop.time() / self.failure_window)
                * self.failure_window
            )

            self._previous_count = self.total_completed
            self._previous_count_succeeded = self.succeeded
            self._previous_count_failed = self.failed

            self.failed = 0
            self.succeeded = 0
            self.total_completed = 0

        self._rate_per_sec = (
            self._previous_count
            * (self.failure_window - (self._loop.time() - self._current_time))
            / self.failure_window
        ) + self.total_completed

        self._rate_per_sec_succeeded = (
            self._previous_count_succeeded
            * (self.failure_window - (self._loop.time() - self._current_time))
            / self.failure_window
        ) + self.succeeded

        self._rate_per_sec_failed = (
            self._previous_count_failed
            * (self.failure_window - (self._loop.time() - self._current_time))
            / self.failure_window
        ) + self.failed

        success_rate = self._rate_per_sec_succeeded / (1 - self.failure_threshold)

        rejection_probability = max(
            (self._rate_per_sec - success_rate) / (self._rate_per_sec + 1), 0
        ) ** (1 / self.rejection_sensitivity)

        return random.random() < rejection_probability

    async def __setup__(self):

        async with self._logger.context() as ctx:
            await ctx.log(Event(
                level=LogLevel.DEBUG,
                message=f'Setting up middleware - {self.__class__.__name__}',
            ))

            self._loop = asyncio.get_event_loop()
            self._current_time = self._loop.time()

    async def __run__(
        self, 
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ) -> MiddlewareResult:
        reject = self.reject_request()

        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            
            await ctx.log(Event(
                level=LogLevel.DEBUG,
                message=f'Request - {context.method} {context.path}:{context.ip_address} - Entered circuit breaker current state - {self._breaker_state.value}',
            ))
            
            if (
                self._breaker_state == CircuitBreakerState.OPEN
                and self._closed_elapsed < self.failure_window
            ):
                self._closed_elapsed = self._loop.time() - self._closed_window_start
                reject = True

                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Not enough time has elapsed since failure state - rejecting request',
                ))

            elif self._breaker_state == CircuitBreakerState.OPEN:
                self._breaker_state = CircuitBreakerState.HALF_OPEN

                self._half_open_window_start = self._loop.time()
                self._closed_elapsed = 0

                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Setting breaker state from {CircuitBreakerState.OPEN.value} to {self._breaker_state.value}',
                ))

            if (
                self._breaker_state == CircuitBreakerState.HALF_OPEN
                and self._half_open_elapsed < self.failure_window
            ):
                self._half_open_elapsed = self._loop.time() - self._half_open_window_start

                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - {self._half_open_elapsed} seconds elapsed since entered {self._breaker_state.value} state',
                ))

            elif self._breaker_state == CircuitBreakerState.HALF_OPEN:
                self._breaker_state = CircuitBreakerState.CLOSED
                self._half_open_elapsed = 0

                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Setting breaker state from {CircuitBreakerState.HALF_OPEN.value} to {self._breaker_state.value}',
                ))

                await ctx.log(Event(
                    level=LogLevel.WARN,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Request tripped circuit breaker',
                ))

            if reject:
                context.response_headers["x-mercury-sync-overload"] = True
                context.status = 503

                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Rejecting request with status of {context.status}',
                ))

            else:
                try:

                    if self.wraps:

                        await ctx.log(Event(
                            level=LogLevel.DEBUG,
                            message=f'Request - {context.method} {context.path}:{context.ip_address} - Executing wrapped middleware {handler.__class__.__name__}',
                        ))

                        (context, response) = await asyncio.wait_for(
                            handler(
                                context=context,
                                response=response
                            ), 
                            timeout=self.handler_timeout
                        )

                        await ctx.log(Event(
                            level=LogLevel.DEBUG,
                            message=f'Request - {context.method} {context.path}:{context.ip_address} - {handler.__class__.__name__} completed middleware',
                        ))

                    else:

                        await ctx.log(Event(
                            level=LogLevel.DEBUG,
                            message=f'Request - {context.method} {context.path}:{context.ip_address} - Executing wrapped request handler',
                        ))

                        response = await asyncio.wait_for(
                            handler(
                                *context.args, 
                                **context.kwargs
                            ), 
                            timeout=self.handler_timeout
                        )

                        await ctx.log(Event(
                            level=LogLevel.DEBUG,
                            message=f'Request - {context.method} {context.path}:{context.ip_address} - Request handler completed execution',
                        ))
                    
                    context.response_headers["x-mercury-sync-overload"] = False

                except (
                    asyncio.TimeoutError,
                    asyncio.CancelledError,
                ):
                    context.response_headers["x-mercury-sync-overload"] = True
                    context.status = 503

                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Request timed out with status of {context.status}',
                    ))

                # Don't count rejections toward failure stats.
                if context.status and context.status >= 400:
                    self.failed += 1

                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Updated failed count {self.failed}',
                    ))

                elif context.status is None or (
                    context.status and context.status < 400
                ):
                    self.succeeded += 1

                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Updated successful count {self.succeeded}',
                    ))

                self.total_completed += 1

                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Updated completed count {self.total_completed}',
                ))

            breaker_open = (
                self._breaker_state == CircuitBreakerState.CLOSED
                or self._breaker_state == CircuitBreakerState.HALF_OPEN
            )

            await ctx.log(Event(
                level=LogLevel.DEBUG,
                message=f'Request - {context.method} {context.path}:{context.ip_address} - Breaker state open is - {breaker_open}',
            ))

            if self.trip_breaker() and breaker_open:
                self._breaker_state = CircuitBreakerState.OPEN
                reject = True

                self._closed_window_start = self._loop.time()
                self._half_open_elapsed = 0

                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Resetting circuit breaker',
                ))

            if reject:
                await ctx.log(Event(
                    level=LogLevel.WARN,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Circuit breaker rejected request',
                ))

                context.errors.append(Exception('Err. - request temporarily rejected.')) 

            return (
                context,
                response,
            ), reject is False

    async def close(self):
        await self._logger.close()

    def abort(self):
        self._logger.abort()


from typing import Any, Dict, List, Literal

from pydantic import BaseModel, StrictBytes, StrictInt, StrictStr

from mkfst.models import (
    HTML,
    Body,
    FileUpload,
    Headers,
    Parameters,
    Query,
    Model,
)

from .parsed_endpoint_metadata import (
    HTTPMethod,
    ParsedEndpointMetadata,
)


class ParsedEndpoint(BaseModel):
    path: StrictStr
    methods: List[HTTPMethod]
    endpoint_metadata: ParsedEndpointMetadata
    responses: Dict[
        StrictInt,
        type[HTML]
        | type[FileUpload]
        | type[Body]
        | type[Model]
        | type[dict]
        | type[list]
        | type[str]
        | type[bytes]
        | StrictStr
        | StrictBytes
        | None,
    ]
    response_headers: Dict[StrictStr, Any] | None = None
    headers: type[Headers] | None = None
    parameters: type[Parameters] | None = None
    query: type[Query] | None = None
    body: (
        type[HTML]
        | type[FileUpload]
        | type[Body]
        | type[Model]
        | type[dict]
        | type[list]
        | type[str]
        | type[bytes]
        | StrictStr
        | StrictBytes
        | None
    ) = None
    required: (
        List[
            Literal[
                "parameters",
                "headers",
                "query",
                "body",
            ]
        ]
        | None
    ) = None

    def to_dict(self):
        return {field: getattr(self, field) for field in self.model_fields}



from base64 import b64decode, b64encode
from http.cookies import BaseCookie, SimpleCookie
from secrets import compare_digest, token_urlsafe
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Set,
)

import zstandard

from mkfst.encryption import AESGCMFernet
from mkfst.env import Env, load_env
from mkfst.logging import Logger, LogLevel
from mkfst.middleware.base import Middleware
from mkfst.middleware.base.response_context import ResponseContext
from mkfst.middleware.base.types import Handler, MiddlewareHandler, MiddlewareResult
from mkfst.models.logging import Event


class CRSF(Middleware):
    def __init__(
        self,
        secret_bytes_size: Optional[int] = 16,
        required_paths: Optional[List[str]] = None,
        exempt_paths: Optional[List[str]] = None,
        sensitive_cookies: Optional[Set[str]] = None,
        safe_methods: List[
            Literal["GET", "HEAD", "OPTIONS", "POST", "PUT", "PATCH", "DELETE", "TRACE"]
        ] = ["GET", "HEAD", "OPTIONS", "TRACE"],
        cookie_name: str = "csrftoken",
        cookie_path: str = "/",
        cookie_domain: Optional[str] = None,
        cookie_secure: bool = False,
        cookie_httponly: bool = False,
        cookie_samesite: str = "lax",
        header_name: str = "x-csrftoken",
    ) -> None:
        env = load_env(Env)

        self.encryptor = AESGCMFernet(env)
        self.secret_bytes_size = secret_bytes_size

        self.required_paths = required_paths
        self.exempt_paths = exempt_paths
        self.sensitive_cookies = sensitive_cookies
        self.safe_methods = safe_methods
        self.cookie_name = cookie_name
        self.cookie_path = cookie_path
        self.cookie_domain = cookie_domain
        self.cookie_secure = cookie_secure
        self.cookie_httponly = cookie_httponly
        self.cookie_samesite = cookie_samesite
        self.header_name = header_name

        self._compressor = zstandard.ZstdCompressor()
        self._decompressor = zstandard.ZstdDecompressor()
        self._logger = Logger()

        super().__init__(self.__class__.__name__, response_headers={})

    async def __run__(
        self, 
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ) -> MiddlewareResult:
        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            
            await ctx.log(Event(
                level=LogLevel.DEBUG,
                message=f'Request - {context.method} {context.path}:{context.ip_address} - Verifying CRSF token'
            ))
            
            (headers, cookies) = context.get_headers_and_cookies()
            request_path = context.path
            request_method = context.method

            crsf_cookie = cookies.get(self.cookie_name)

            is_unsafe_method = request_method not in self.safe_methods

            path_is_required = False
            if self.required_paths:
                path_is_required = self._path_is_required(request_path)

            path_is_exempt = False
            if self.exempt_paths:
                path_is_exempt = self._path_is_exempt(request_path)

            has_sensitive_cookies = False
            if self.sensitive_cookies:
                has_sensitive_cookies = self._has_sensitive_cookies(cookies)

            is_sensitive = is_unsafe_method and not path_is_exempt and has_sensitive_cookies

            if path_is_required or is_sensitive:
                submitted_csrf_token: str | None = headers.get(self.header_name)

                csrf_tokens_match = False

                try:
                    decoded_crsf_cookie: str = self.encryptor.decrypt(
                        self._decompressor.decompress(b64decode(crsf_cookie.encode()))
                    )
                    decoded_crsf_token: str = self.encryptor.decrypt(
                        self._decompressor.decompress(
                            b64decode(submitted_csrf_token.encode())
                        )
                    )

                    csrf_tokens_match = compare_digest(
                        decoded_crsf_cookie, decoded_crsf_token
                    )

                except Exception:
                    await ctx.log(Event(
                        level=LogLevel.ERROR,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Rejecting request and could not verify CRSF token'
                    ))

                    csrf_tokens_match = False

                crsf_match_failed = (
                    crsf_cookie is None
                    or submitted_csrf_token is None
                    or csrf_tokens_match is False
                )

                if crsf_match_failed:
                    context.status = 403
                    await ctx.log(Event(
                        level=LogLevel.ERROR,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Rejecting request due to invalid CRSF token'
                    ))

                    return (
                        context,
                        "CSRF token verification failed",
                    ), False

            crsf_cookie = cookies.get(self.cookie_name)

            response_headers = {}

            if crsf_cookie is None:

                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Creating CRSF token'
                ))
                
                cookie: BaseCookie = SimpleCookie()
                cookie_name = self.cookie_name

                crsf_token = self.encryptor.encrypt(
                    token_urlsafe(nbytes=self.secret_bytes_size).encode()
                )

                cookie[cookie_name] = b64encode(
                    self._compressor.compress(crsf_token)
                ).decode()

                cookie[cookie_name]["path"] = self.cookie_path
                cookie[cookie_name]["secure"] = self.cookie_secure
                cookie[cookie_name]["httponly"] = self.cookie_httponly
                cookie[cookie_name]["samesite"] = self.cookie_samesite

                if self.cookie_domain is not None:
                    cookie[cookie_name]["domain"] = self.cookie_domain  # pragma: no cover

                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Adding CRSF token to headers'
                ))

                response_headers["set-cookie"] = cookie.output(header="").strip()

            context.response_headers.update(response_headers)

            await ctx.log(Event(
                level=LogLevel.DEBUG,
                message=f'Request - {context.method} {context.path}:{context.ip_address} - Request met CRSF validation requirements'
            ))

            return (
                context,
                response,
            ), True

    def _has_sensitive_cookies(self, cookies: Dict[str, str]) -> bool:
        for sensitive_cookie in self.sensitive_cookies:
            if cookies.get(sensitive_cookie) is not None:
                return True

        return False

    def _path_is_required(self, path: str) -> bool:
        for required_url in self.required_paths:
            if required_url in path:
                return True

        return False

    def _path_is_exempt(self, path: str) -> bool:
        for exempt_path in self.exempt_paths:
            if exempt_path in path:
                return True

        return False



import asyncio

from mkfst import (
    Service,
    endpoint,
    Parameters,
    Env,
)


class TestParams(Parameters):
    id: str


class TestService(Service):
    @endpoint("/")
    async def get_service(self) -> str:
        return "Hello World"

    # @endpoint(
    #     "/post",
    #     methods=["POST"],
    # )
    # async def post_data(self, data: dict) -> dict:
    #     return data

    # @endpoint("/get/:id")
    # async def get_by_id(self, params: TestParams) -> str:
    #     return params.id


if __name__ == "__main__":

    async def run_server():
        server = TestService(
            "localhost",
            6099,
            log_level="error",
            workers=18,
            env=Env(MERCURY_SYNC_CLEANUP_INTERVAL="0.25s"),
        )

        await server.run()

    asyncio.run(run_server())



from __future__ import annotations
from enum import Enum
from typing import Literal

LogLevelName = Literal["trace", "debug", "info", "warn", "error", "critical", "fatal"]


class LogLevel(Enum):
    TRACE = "TRACE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARN = "WARN"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"
    FATAL = "FATAL"

    @classmethod
    def to_level(cls, level_name: LogLevelName) -> LogLevel:
        levels_map = {
            LogLevel.TRACE.value: LogLevel.TRACE,
            LogLevel.DEBUG.value: LogLevel.DEBUG,
            LogLevel.INFO.value: LogLevel.INFO,
            LogLevel.WARN.value: LogLevel.WARN,
            LogLevel.ERROR.value: LogLevel.ERROR,
            LogLevel.CRITICAL.value: LogLevel.CRITICAL,
            LogLevel.FATAL.value: LogLevel.FATAL,
        }

        return levels_map.get(level_name.upper())



import asyncio
from typing import (
    AsyncGenerator,
    Callable,
    TypeVar,
)

from mkfst.logging.models import Log

from .consumer_status import ConsumerStatus

T = TypeVar("T")


class LogConsumer:
    def __init__(self) -> None:
        self._queue: asyncio.Queue[Log] = asyncio.Queue()
        self._wait_task: asyncio.Task | None = None
        self._loop = asyncio.get_event_loop()
        self._pending_waiter: asyncio.Future | None = None
        self._yield_lock = asyncio.Lock()
        self.status = ConsumerStatus.READY

    @property
    def pending(self):
        return self._queue.qsize() > 0

    async def wait_for_pending(self):
        if self.status == ConsumerStatus.CLOSING:
            self._pending_waiter = asyncio.Future()
            await self._pending_waiter

    async def iter_logs(
        self,
        filter: Callable[[T], bool] | None = None,
    ) -> AsyncGenerator[Log, None]:
        if self.status == ConsumerStatus.READY:
            self.status = ConsumerStatus.RUNNING

        try:
            while self.status == ConsumerStatus.RUNNING:
                self._wait_task = asyncio.create_task(self._queue.get())

                log: Log = await self._wait_task

                if filter and filter(log.entry):
                    yield log

                elif filter is None:
                    yield log

                else:
                    self._queue.put_nowait(log)

        except (asyncio.CancelledError, asyncio.InvalidStateError):
            pass

        remaining = self._queue.qsize()

        if self.status == ConsumerStatus.CLOSING:
            for _ in range(remaining):
                self._wait_task = asyncio.create_task(self._queue.get())
                log: Log = await self._wait_task

                if filter and filter(log.entry):
                    yield log

                elif filter is None:
                    yield log

        if self._pending_waiter and not self._pending_waiter.done():
            self._pending_waiter.set_result(None)

        self.status = ConsumerStatus.CLOSED

    async def put(self, log: Log):
        await self._queue.put(log)

    def abort(self):
        self.status = ConsumerStatus.ABORTING
        if self._wait_task:
            try:
                self._wait_task.cancel()

            except asyncio.CancelledError:
                pass

            except asyncio.InvalidStateError:
                pass

        remaining = self._queue.qsize()
        for _ in range(remaining):
            self._queue.get_nowait()

        self.status = ConsumerStatus.CLOSED

    def stop(self):
        self.status = ConsumerStatus.CLOSING

        if self._queue.qsize() < 1 and self._wait_task:
            try:
                self._wait_task.cancel()

            except (
                asyncio.CancelledError,
                asyncio.InvalidStateError,
            ):
                pass



import functools
from typing import Literal, Optional

from mkfst.env.time_parser import TimeParser


def task(
    schedule: Optional[float | int | str]=None,
    trigger: Literal['MANUAL', 'ON_START']='MANUAL',
    repeat: Literal['NEVER', 'ALWAYS'] | int='NEVER',
    timeout: Optional[int | float | str]=None,
    keep: Optional[int]=10,
    max_age: Optional[str]='10m',
    keep_policy: Literal['COUNT', 'AGE', 'COUNT_AND_AGE']='COUNT'
):
    
    if isinstance(schedule, str):
        parser = TimeParser(schedule)
        schedule = parser.time

    if isinstance(timeout, str):
        parser = TimeParser(timeout)
        timeout = parser.time

    
    if schedule is None and repeat == 'ALWAYS':
        
        raise Exception('Err. - A schedule must be specified if repeat is set as "ALWAYS" ')
    
    elif schedule is None and isinstance(repeat, int) and repeat > 0:
        raise Exception('Err. - A non-zero number of repeats must be specified if repeat is set as "ALWAYS" ')
    
    if not isinstance(keep, int) and keep_policy == 'COUNT':
        raise Exception('Err. - cannot have null or non-zero keep count if keep_policy is "COUNT')
    
    elif not isinstance(max_age, str) and keep_policy == 'AGE':
        raise Exception('Err. - cannot have null max_age if keep_policy is "AGE')
    
    if max_age:
        max_age = TimeParser(max_age).time

    def wraps(func):

        func.name = func.__name__
        func.schedule = schedule
        func.trigger = trigger
        func.repeat = repeat
        func.as_task = True
        func.timeout = timeout
        func.keep = keep
        func.max_age = max_age
        func.keep_policy = keep_policy

        @functools.wraps(func)
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)
        
        return decorator
    
    return wraps


from pydantic import AnyUrl, BaseModel, StrictStr

from .models import EmailStr


class ParsedAPIMetadata(BaseModel):
    title: StrictStr
    version: StrictStr
    summary: StrictStr | None = None
    terms_of_service: StrictStr | None = None
    description: StrictStr | None = None
    owner: StrictStr | None = None
    owner_url: AnyUrl | None = None
    owner_email: EmailStr | None = None
    license: StrictStr | None = None
    license_identifier: StrictStr | None = None
    license_url: AnyUrl | None = None

    class Config:
        arbitrary_types_allowed = True


from .cors import Cors as Cors
from .crsf import CRSF as CRSF

from .circuit_breaker import CircuitBreaker as CircuitBreaker

from .compressor import (
    BidirectionalGZipCompressor as BidirectionalGZipCompressor,
    BidirectionalZStandardCompressor as BidirectionalZStandardCompressor,
    GZipCompressor as GZipCompressor,
    ZStandardCompressor as ZStandardCompressor,
)

from .connection import (
    HTTPSRedirect as HTTPSRedirect,
    TrustedHost as TrustedHost,
)

from .decompressor import (
    BidirectionalGZipDecompressor as BidirectionalGZipDecompressor,
    BidirectionalZStandardDecompressor as BidirectionalZStandardDecompressor,
    GZipDecompressor as GZipDecompressor,
    ZStandardDecompressor as ZStandardDecompressor,
)



from .models import Entry as Entry
from .models import LogLevel as LogLevel
from .streams import Logger as Logger
from .config import LoggingConfig as LoggingConfig
from .models import LogLevelName as LogLevelName


import os
import msgspec
from typing import Dict, Type, TypeVar, Union

from dotenv import dotenv_values

from .env import Env

T = TypeVar("T", bound=msgspec.Struct)

PrimaryType = Union[str, int, bool, float, bytes]


def load_env(default: type[Env], env_file: str = None, override: T | None = None) -> T:
    envars = default.types_map()

    if env_file is None:
        env_file = ".env"

    values: Dict[str, PrimaryType] = {}
    for envar_name, envar_type in envars.items():
        envar_value = os.getenv(envar_name)
        if envar_value:
            values[envar_name] = envar_type(envar_value)

    if env_file and os.path.exists(env_file):
        env_file_values = dotenv_values(dotenv_path=env_file)

        for envar_name, envar_value in env_file_values.items():
            envar_type = envars.get(envar_name)
            if envar_type:
                env_file_values[envar_name] = envar_type(envar_value)

        values.update(env_file_values)

    if override:
        values.update(**msgspec.structs.asdict(override))

        return type(override)(
            **{name: value for name, value in values.items() if value is not None}
        )

    return default(
        **{name: value for name, value in values.items() if value is not None}
    )



from pydantic import AnyHttpUrl, BaseModel, StrictStr

from .models import ExternalDocumentation


class ParsedTag(BaseModel):
    value: StrictStr
    description: StrictStr | None = None
    docs_description: StrictStr | None = None
    docs_url: AnyHttpUrl | None = None

    def parse(self):

        tag = {
            "name": self.value,
            "description": self.description,
        }

        docs: ExternalDocumentation | None = None
        if self.docs_url:
            docs = {
                "description": self.docs_description,
                "url": self.docs_url.unicode_string()
            }
            
            tag["externalDocs"] = docs

        return tag


from __future__ import annotations
import time
import msgspec
from typing import Literal, Any, Dict, Tuple
from .run_status import RunStatus
from .task_type import TaskType


CommandType = Literal["shell", "subprocess"]


class ShellProcess(msgspec.Struct):
    run_id: int
    task_name: str
    process_id: int
    command: str
    status: RunStatus
    args: Tuple[str, ...] | None = None
    return_code: int | None = None
    env: Dict[str, Any] | None = None
    working_directory: str | None = None
    command_type: CommandType = "subprocess"
    error: str | None = None
    trace: str | None = None
    start: int | float = time.monotonic()
    end: int | float | None = None
    elapsed: int | float = 0
    result: str | None = None
    task_type: TaskType = TaskType.SHELL

    def complete(self):
        return self.status in [
            RunStatus.COMPLETE,
            RunStatus.CANCELLED,
            RunStatus.FAILED,
        ]



from typing import Dict, List

from pydantic import AnyHttpUrl, BaseModel, StrictStr


class ParsedServerVariable(BaseModel):
    options: List[StrictStr] | None = None
    default: StrictStr
    description: StrictStr | None = None


class ParsedAPIServer(BaseModel):
    server_url: AnyHttpUrl
    server_description: StrictStr | None
    server_variables: Dict[StrictStr, ParsedServerVariable] | None


from __future__ import annotations

import inspect
from collections import defaultdict
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    List,
    Literal,
    Tuple,
    get_args,
    get_type_hints,
    get_origin,
)

from mkfst.connection.tcp.fabricator import Fabricator
from mkfst.docs import (
    ParsedEndpoint,
    ParsedEndpointMetadata,
    ParsedOperationMetadata,
    ParsedTag,
)
from mkfst.env import Env
from mkfst.middleware.base import Middleware
from mkfst.middleware.base.base_wrapper import BaseWrapper
from mkfst.models.http import (
    Model,
)
from mkfst.tasks import TaskRunner
from .parse_to_source_type import parse_to_source_type


Json = dict | list

Tag = Dict[Literal["value", "description", "docs_description", "docs_url"], str]


def join_paths(*paths: str):
    joined = "/".join([path.strip("/") for path in paths])

    return f"/{joined}"


class Group:
    def __init__(
        self,
        base: str,
        groups: List[Group] | None = None,
        middleware: List[Middleware] | None = None,
        tags: List[Tag] | None = None,
    ) -> None:
        if groups is None:
            groups = []

        if middleware is None:
            middleware = []

        self.service_name = self.__class__.__name__
        self._base = base
        self._groups = groups
        self._middleware: List[Middleware] = middleware
        self._supported_handlers: Dict[str, Dict[str, str]] = defaultdict(dict)
        self._handlers: Dict[str, Callable[..., Awaitable[Any]]] = {}

        self.tags = tags or []

        if self.service_name != "Group":
            self.tags.append(
                ParsedTag(
                    value=self.service_name,
                    description=f"The {self.service_name} service group.",
                )
            )

        if groups:
            for group in groups:
                self.tags.extend(group.tags)

    def _assemble(
        self,
        instance_id: int,
        env: Env,
        parent_middleware: List[Middleware],
    ):
        self._middleware.extend(
            [
                middleware
                for middleware in parent_middleware
                if middleware not in self._middleware
            ]
        )

        task_runner = TaskRunner(instance_id, env)

        parsers: Dict[str, Any] = {}
        events: Dict[str, Dict[str, Callable[..., Awaitable[Any]]]] = {}
        match_routes: Dict[str, Dict[str, Callable[..., Awaitable[Any]]]] = {}

        request_parsers: Dict[str, Model | dict | list | str] = {}
        response_parsers: Dict[Model, Tuple[Callable[[Any], str], int]] = {}

        middleware_enabled: Dict[str, bool] = {}

        routes: Dict[str, Dict[str, Callable[..., Awaitable[Any]]]] = {}

        (
            endpoints,
            tasks,
            fabricators,
            endpoint_docs,
            response_headers,
        ) = self._gather_hooks()

        for task in tasks.values():
            task_runner.add(task)

        for path, endpoint in endpoints.items():
            has_middleware = len(self._middleware) > 0
            if has_middleware:
                middleware_enabled[path] = True

            if isinstance(endpoint, (Middleware, BaseWrapper)):
                endpoint = self._handlers[path]

            endpoint_signature = inspect.signature(endpoint)
            params = endpoint_signature.parameters.values()

            return_type = get_type_hints(endpoint).get("return")
            response_types = get_args(return_type)

            methods = endpoint.methods

            request_parsers.update(
                {
                    join_paths(self._base, endpoint.path): {
                        method: get_args(param_type.annotation)[0]
                    }
                    for param_type in params
                    for method in methods
                    if (
                        (args := get_args(param_type.annotation))
                        and len(args) > 0
                        and args[0] in Model.__subclasses__()
                    )
                }
            )

            routes[join_paths(self._base, endpoint.path)] = {
                method: endpoint for method in endpoint.methods
            }

            if (
                len(response_types) > 1
                and inspect.isclass(response_types[0])
                and response_types[0] in Model.__subclasses__()
            ):
                model = response_types[0]
                status_code = response_types[1]

                response_parsers.update(
                    {
                        join_paths(self._base, endpoint.path): {
                            method: (
                                model,
                                status_code,
                            )
                            for method in methods
                        }
                    }
                )

            elif (
                len(response_types) > 0 and response_types[0] in Model.__subclasses__()
            ):
                model = response_types[0]

                response_parsers.update(
                    {
                        join_paths(self._base, endpoint.path): {
                            method: (
                                model,
                                200,
                            )
                            for method in methods
                        }
                    }
                )

            elif return_type in Model.__subclasses__():
                response_parsers.update(
                    {
                        join_paths(self._base, endpoint.path): {
                            method: (
                                return_type,
                                200,
                            )
                            for method in methods
                        }
                    }
                )

            elif return_type in get_args(Json) or get_origin(return_type) in get_args(
                Json
            ):
                response_parsers.update(
                    {
                        join_paths(self._base, endpoint.path): {
                            method: (
                                parse_to_source_type(
                                    return_type,
                                ),
                                200,
                            )
                            for method in methods
                        }
                    }
                )

            if isinstance(endpoint.responses, dict):
                responses = endpoint.responses

                response_parsers.update(
                    {
                        join_paths(self._base, endpoint.path): {
                            method: (response_model, status)
                            for method in methods
                            for status, response_model in responses.items()
                            if (issubclass(response_model, Model))
                        }
                    }
                )

        parsers.update(request_parsers)
        parsers.update(response_parsers)

        events.update(
            {
                join_paths(
                    self._base,
                    path,
                ): endpoint
                for path, endpoint in endpoints.items()
            }
        )
        match_routes.update(routes)

        for group in self._groups:
            group._base = join_paths(self._base, group._base)

        assembled_groups = self._gather_groups(
            instance_id,
            env,
        )

        group_middleware = list(self._middleware)

        for assembled in assembled_groups:
            routes.update(assembled["routes"])
            match_routes.update(assembled["match_routes"])
            events.update(assembled["events"])
            parsers.update(assembled["parsers"])
            middleware_enabled.update(assembled["middleware_enabled"])
            self._supported_handlers.update(assembled["supported_handlers"])
            request_parsers.update(assembled["request_parsers"])
            response_parsers.update(assembled["response_parsers"])
            fabricators.update(assembled["fabricators"])
            endpoint_docs.update(assembled["endpoint_docs"])
            response_headers.update(assembled["response_headers"])
            group_middleware.extend(
                [
                    middleware
                    for middleware in assembled["middleware"]
                    if middleware not in self._middleware
                ]
            )

        return {
            "routes": routes,
            "match_routes": match_routes,
            "events": events,
            "parsers": parsers,
            "middleware_enabled": middleware_enabled,
            "supported_handlers": self._supported_handlers,
            "request_parsers": request_parsers,
            "response_parsers": response_parsers,
            "fabricators": fabricators,
            "endpoint_docs": endpoint_docs,
            "middleware": group_middleware,
            "response_headers": response_headers,
        }

    def _gather_hooks(self):
        reserved = ["connect", "close"]
        endpoints: Dict[
            str, Callable[..., Awaitable[Model | Dict[Any, Any] | str]]
        ] = {}
        tasks: Dict[str, Callable[[], Awaitable[Any]]] = {}
        fabricators: Dict[str, Fabricator] = {}
        endpoint_docs: Dict[str, ParsedEndpoint] = {}
        response_headers: Dict[str, Dict[str, Any]] = {}

        for _, call in inspect.getmembers(self, predicate=inspect.ismethod):
            hook_name: str = call.__name__
            not_internal = hook_name.startswith("__") is False
            not_reserved = hook_name not in reserved
            is_endpoint = hasattr(call, "as_endpoint")
            is_task = hasattr(call, "as_task")

            if not_internal and not_reserved and is_endpoint:
                methods: List[str] = call.methods

                handler = call
                for middleware_operator in self._middleware:
                    call = middleware_operator.wrap(call)

                endpoints.update({handler.path: call})

                self._handlers.update({handler.path: handler})

                call_params = inspect.signature(handler).parameters

                required_params = [
                    (key, value.annotation)
                    for key, value in call_params.items()
                    if value.default == inspect._empty
                ]

                optional_params = {
                    key: get_args(value.annotation)
                    for key, value in call_params.items()
                    if value.default != inspect._empty
                }

                fabricator = Fabricator(required_params, optional_params)

                return_type = get_type_hints(handler).get("return")
                response_types = get_args(return_type)

                responses: Dict[int, Any] = {}

                if (
                    len(response_types) > 1
                    and inspect.isclass(response_types[0])
                    and response_types[0] in Model.__subclasses__()
                ):
                    model = response_types[0]
                    status_code = response_types[1]

                    responses[status_code] = model

                elif len(response_types) == 1:
                    responses[200] = response_types[0]

                elif (
                    len(response_types) > 0
                    and response_types[0] in Model.__subclasses__()
                ):
                    model = response_types[0]
                    responses[200] = model

                else:
                    responses[200] = return_type

                if isinstance(handler.responses, dict):
                    responses.update(handler.responses)

                additional_docs: Dict[Literal["docs_description", "docs_url"], str] = {}

                if isinstance(handler.additional_docs, dict):
                    additional_docs = handler.additional_docs

                method_metadata: Dict[
                    Literal[
                        "GET",
                        "HEAD",
                        "OPTIONS",
                        "POST",
                        "PUT",
                        "PATCH",
                        "DELETE",
                        "TRACE",
                    ],
                    Dict[
                        Literal[
                            "name",
                            "tags",
                            "description",
                            "summary",
                            "docs_url",
                            "docs_description",
                            "depreciated",
                        ],
                        str | List[str] | bool,
                    ],
                ] = handler.method_metadata

                call_response_headers = dict(handler.response_headers)
                response_headers.update(
                    {
                        handler.path: {
                            method: dict(handler.response_headers) for method in methods
                        }
                    }
                )

                endpoint_docs[handler.path] = ParsedEndpoint(
                    path=handler.path,
                    methods=methods,
                    endpoint_metadata=ParsedEndpointMetadata(
                        description=handler.__doc__,
                        summary=handler.summary,
                        operations={
                            method: ParsedOperationMetadata(
                                group_name=self.service_name,
                                name=method_metadata[method].get(
                                    "name", handler.__name__
                                ),
                                tags=method_metadata[method].get("tags"),
                                description=method_metadata[method].get("description"),
                                docs_description=additional_docs.get(
                                    "docs_description"
                                ),
                                docs_url=additional_docs.get("docs_url"),
                                deprecated=method_metadata[method].get("depreciated"),
                            )
                            for method in methods
                        },
                    ),
                    responses={
                        status: parse_to_source_type(response_type)
                        for status, response_type in responses.items()
                    },
                    response_headers=call_response_headers,
                    headers=fabricator["headers"],
                    parameters=fabricator["parameters"],
                    query=fabricator["query"],
                    body=fabricator["body"],
                    required=fabricator.required_params,
                )

                fabricators[join_paths(self._base, handler.path)] = {
                    method: fabricator for method in methods
                }

                self._supported_handlers[join_paths(self._base, handler.path)] = {
                    method: handler for method in methods
                }

            if not_internal and not_reserved and is_task:
                hook_name = handler.__name__
                tasks[hook_name] = handler

        return (
            endpoints,
            tasks,
            fabricators,
            endpoint_docs,
            response_headers,
        )

    def _gather_groups(
        self, instance_id: int, env: Env
    ) -> List[
        Dict[
            Literal[
                "routes",
                "match_routes",
                "events",
                "parsers",
                "middleware_enabled",
                "supported_handlers",
                "response_parsers",
                "request_parsers",
                "fabricators",
                "endpoint_docs",
                "middleware",
                "response_headers",
            ],
            Dict[str, Any],
        ]
    ]:
        return [
            group._assemble(instance_id, env, self._middleware)
            for group in self._groups
        ]



from __future__ import annotations
import msgspec
from typing import Dict, Literal


class CorsHeaders(msgspec.Struct, kw_only=True):
    access_control_allow_origin: list[str]
    access_control_expose_headers: list[str] | None = None
    access_control_max_age: int | float | None = None
    access_control_allow_credentials: bool | None = None
    access_control_allow_methods: list[
        Literal["GET", "HEAD", "OPTIONS", "POST", "PUT", "PATCH", "DELETE", "TRACE"],
    ]
    access_control_allow_headers: list[str] | None = None

    def to_headers(self):
        cors_headers: Dict[str, str] = {}

        headers = self.dict(exclude_none=True)

        for key, value in headers.items():
            header_key = "-".join([segment.capitalize() for segment in key.split("_")])

            if key == "access_control_allow_origin":
                header_value = " | ".join(value)

            elif key == "access_control_max_age":
                header_value = "true" if value else "false"

            else:
                header_value = ", ".join(value)

            cors_headers[header_key] = header_value

        return cors_headers

    def to_simple_headers(self):
        allow_all_origins = False
        allow_all_origins = "*" in self.access_control_allow_origin
        simple_headers: Dict[str, str] = {}

        if allow_all_origins:
            simple_headers["Access-Control-Allow-Origin"] = "*"

        if self.access_control_allow_credentials:
            simple_headers["Access-Control-Allow-Credentials"] = "true"

        if self.access_control_expose_headers:
            simple_headers["Access-Control-Expose-Headers"] = ", ".join(
                self.access_control_expose_headers
            )

        return simple_headers

    def to_preflight_headers(self):
        allow_all_origins = "*" in self.access_control_allow_origin

        access_control_allow_headers = self.access_control_allow_headers or []
        allow_all_headers = "*" in access_control_allow_headers

        safe_headers = {"Accept", "Accept-Language", "Content-Language", "Content-Type"}

        preflight_explicit_allow_origin = (
            not allow_all_origins or self.access_control_allow_credentials
        )

        preflight_headers: Dict[str, str] = {}
        if preflight_explicit_allow_origin:
            # The origin value will be set in preflight_response() if it is allowed.
            preflight_headers["Vary"] = "Origin"

        else:
            preflight_headers["Access-Control-Allow-Origin"] = "*"

        preflight_headers.update(
            {
                "Access-Control-Allow-Methods": ", ".join(
                    self.access_control_allow_methods
                ),
                "Access-Control-Max-Age": str(self.access_control_max_age),
            }
        )

        allow_headers = sorted(safe_headers | set(access_control_allow_headers))

        if allow_headers and not allow_all_headers:
            preflight_headers["Access-Control-Allow-Headers"] = ", ".join(allow_headers)

        if self.access_control_allow_credentials:
            preflight_headers["Access-Control-Allow-Credentials"] = "true"

        return preflight_headers



from base64 import b64decode
from typing import (
    Any,
    Dict,
)

import zstandard

from mkfst.logging import Logger, LogLevel
from mkfst.middleware.base import Middleware, MiddlewareType
from mkfst.middleware.base.response_context import ResponseContext
from mkfst.middleware.base.types import Handler, MiddlewareHandler, MiddlewareResult
from mkfst.models.http.parse_response import parse_response
from mkfst.models.logging import Event


class BidirectionalZStandardDecompressor(Middleware):
    def __init__(self) -> None:
        super().__init__(
            self.__class__.__name__, 
            middleware_type=MiddlewareType.BIDIRECTIONAL
        )

        self._decompressor = zstandard.ZstdDecompressor()
        self._logger = Logger()

    async def __pre__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ) -> MiddlewareResult:
        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            
            await ctx.log(Event(
                level=LogLevel.DEBUG,
                message=f'Request - {context.method} {context.path}:{context.ip_address} - Decompressing request via ZStd'
            ))
            
            decompressed_data = b''
            try:
                headers_model = context.get_headers()
                data = context.get_bytes_arg()

                headers: Dict[str, Any] = {}
                if headers_model:
                    headers = headers_model.model_dump()

                content_encoding = headers.get(
                    "content-encoding", 
                    context.request_headers.get("x-compression-encoding")
                )

                if data != b"" and content_encoding == "zstd":
                    decompressed_data = self._decompressor.decompress(
                        data
                    )

                    context.update_request_headers({
                        key: value
                        for key, value in headers.items()
                        if key != "content-encoding" and key != "x-compression-encoding"
                    })

                    context.update_request_data(decompressed_data)

                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Decompressed request via ZStd'
                    ))
                
                else:
                    await ctx.log(Event(
                        level=LogLevel.WARN,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - No request to decompress'
                    ))

                return (
                    context,
                    response
                ), True

            except Exception as e:
                context.errors.append(e)
                context.status = 500

                await ctx.log(Event(
                    level=LogLevel.ERROR,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Encountered error ZStd decompressing request - {str(e)}'
                ))

                return (
                    context,
                    response
                ), False

    async def __post__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ) -> MiddlewareResult:
        
        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            
            try:
                content_encoding = context.response_headers.get(
                    "content-encoding", 
                    context.response_headers.get("x-compression-encoding")
                )
                
                if response is None:
                    return (
                        context,
                        response
                    ), True
                
                elif content_encoding != "zstd":
                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - No content-encoding header - skipping decompression'
                    ))

                    return (
                        context,
                        response
                    ), True

                
                if isinstance(response, str):
                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Decompressing string response via ZStd'
                    ))

                    decompressed_data = self._decompressor.decompress(response.encode())

                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Decompressed response via ZStd'
                    ))

                else:
                    
                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Decompressing response via ZStd'
                    ))

                    serialized: str = parse_response(
                        response,
                        context.parser
                    )
                    decompressed_data = self._decompressor.decompress(
                        b64decode(serialized.encode())
                    )

                    context.response_headers.pop(
                        "content-encoding", 
                        context.response_headers.pop("x-compression-encoding")
                    )

                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Decompressed response via ZStd with level'
                    ))

                return (
                    context,
                    decompressed_data.decode()
                ), True

            except Exception as e:
                context.errors.append(e)
                context.status = 500

                await ctx.log(Event(
                    level=LogLevel.ERROR,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Encountered error ZStd decompressing request - {str(e)}'
                ))

                return (
                    context,
                    response
                ), False



import asyncio
import math
import os
import statistics
from typing import List, Union

import psutil

from mkfst.models.http import Limit

from .base_limiter import BaseLimiter


class ResourceAdaptiveLimiter(BaseLimiter):
    __slots__ = (
        "max_rate",
        "time_period",
        "_rate_per_sec",
        "_level",
        "_waiters",
        "_loop",
        "_last_check",
        "_cpu_limit",
        "_current_time",
        "_previous_count",
        "_current_cpu",
        "_max_queue",
        "_sample_task",
        "_running",
        "_process",
        "_max_fast_backoff",
        "_min_backoff",
        "_cpu_history",
        "_memory_history",
        "_memory_limit",
        "_current_memory",
    )

    def __init__(self, limit: Limit) -> None:
        super().__init__(
            limit.max_requests, limit.period, reject_requests=limit.reject_requests
        )

        cpu_limit = limit.cpu_limit
        if cpu_limit is None:
            cpu_limit = 50

        self._cpu_limit = cpu_limit
        self._backoff = limit.backoff
        self._min_backoff = self._backoff
        self._max_fast_backoff = math.ceil(self._backoff * 10)
        self._max_backoff = math.ceil(self._max_fast_backoff * 10)
        self._last_check = self._loop.time()
        self._current_time = self._loop.time()
        self._previous_count = limit.max_requests

        self._memory_limit = limit.memory

        self._cpu_history: List[float] = []
        self._memory_history: List[float] = []

        self._max_queue = limit.max_requests
        self._sample_task: Union[asyncio.Task, None] = None
        self._running = False
        self._activate_limit = False
        self._process = psutil.Process(os.getpid())

        self._current_cpu = self._process.cpu_percent()
        self._current_memory = self._get_memory()

        self._cpu_history.append(self._current_cpu)

    async def acquire(
        self,
        amount: float = 1,
    ):
        if not self._running:
            self._running = True
            self._sample_task = asyncio.create_task(self._sample_cpu())

        if amount > self.max_rate:
            raise ValueError("Can't acquire more than the maximum capacity")

        task = asyncio.current_task(loop=self._loop)

        assert task is not None

        rejected = False

        while self._activate_limit:
            fut = self._loop.create_future()
            try:
                self._waiters[task] = fut

                await asyncio.wait_for(asyncio.shield(fut), timeout=self._backoff)

                self._max_fast_backoff = min(
                    self._max_fast_backoff + (1 / math.sqrt(self._rate_per_sec)),
                    self._max_backoff,
                )

            except asyncio.TimeoutError:
                pass

            fut.cancel()

            rejected = True

        self._backoff = min(self._backoff * 2, self._max_fast_backoff)
        self._waiters.pop(task, None)
        self._level += amount

        return rejected

    async def _sample_cpu(self):
        while self._running:
            self._current_cpu = self._process.cpu_percent()
            self._current_memory = self._get_memory()

            self._cpu_history.append(self._current_cpu)
            self._memory_history.append(self._current_memory)

            elapsed = self._loop.time() - self._last_check

            if elapsed > self.time_period:
                self._cpu_history.pop(0)

            median_cpu_usage = statistics.median(self._cpu_history)
            median_memory_usage = statistics.median(self._memory_history)

            if (
                self._current_cpu >= self._cpu_limit
                or self._current_memory >= self._memory_limit
            ):
                self._activate_limit = True

            elif (
                median_cpu_usage < self._cpu_limit
                and median_memory_usage < self._memory_limit
            ):
                self._activate_limit = False
                self._max_fast_backoff = max(
                    self._max_fast_backoff - (1 / self._rate_per_sec), self._min_backoff
                )

            await asyncio.sleep(0.1)

    def _get_memory(self):
        return self._process.memory_info().rss / 1024**2

    async def close(self):
        self._running = False

        self._sample_task.cancel()
        if not self._sample_task.cancelled():
            try:
                await self._sample_task

            except (asyncio.CancelledError, asyncio.InvalidStateError):
                pass



import contextvars
from typing import List, Literal

from mkfst.logging.models import LogLevel, LogLevelName
from .log_level_map import LogLevelMap
from .stream_type import StreamType


LogOutput = Literal["stdout", "stderr"]

_global_log_level = contextvars.ContextVar("_global_log_level", default=LogLevel.INFO)
_global_disabled_loggers = contextvars.ContextVar(
    "_global_disabled_loggers", default=[]
)
_global_level_map = contextvars.ContextVar("_global_level_map", default=LogLevelMap())
_global_log_output_type = contextvars.ContextVar(
    "_global_log_level_type", default=StreamType.STDOUT
)
_global_logging_directory = contextvars.ContextVar(
    "_global_logging_directory", default=None
)


class LoggingConfig:
    def __init__(
        self,
        level: LogLevelName | None = None,
    ) -> None:
        self._log_level: contextvars.ContextVar[LogLevel] = _global_log_level

        if level:
            self._log_level.set(LogLevel.to_level(level))

        self._log_output_type: contextvars.ContextVar[StreamType] = (
            _global_log_output_type
        )
        self._log_directory: contextvars.ContextVar[str | None] = (
            _global_logging_directory
        )

        self._disabled_loggers: contextvars.ContextVar[List[str]] = (
            _global_disabled_loggers
        )
        self._level_map = _global_level_map.get()

        self._level = _global_log_level.get()
        if level:
            self._level = LogLevel.to_level(level)

    def update(
        self,
        log_directory: str | None = None,
        log_level: LogLevelName | None = None,
        log_output: LogOutput | None = None,
    ):
        if log_directory:
            self._log_directory.set(log_directory)

        if log_level:
            self._log_level.set(LogLevel.to_level(log_level))

        if log_output:
            self._log_output_type.set(
                StreamType.STDOUT if log_output == "stdout" else StreamType.STDERR
            )

    def enabled(self, logger_name: str, log_level: LogLevel) -> bool:
        disabled_loggers = self._disabled_loggers.get()

        current_log_level = self._level
        if current_log_level is None:
            current_log_level = self._log_level.get()

        return logger_name not in disabled_loggers and (
            self._level_map[log_level] >= self._level_map[current_log_level]
        )

    @property
    def level(self):
        return self._log_level.get()

    @property
    def output(self):
        return self._log_output_type.get()

    @property
    def directory(self):
        return self._log_directory.get()



import asyncio
from typing import Optional, OrderedDict

from mkfst.caching.base import BaseCache
from mkfst.caching.serializers import NullSerializer


class SimpleMemoryCache(BaseCache[str]):
    """
    Memory cache implementation with the following components as defaults:
        - serializer: :class:`aiocache.serializers.NullSerializer`
        - plugins: None
        - backend: dict

    Config options are:

    :param serializer: obj derived from :class:`aiocache.serializers.BaseSerializer`.
    :param plugins: list of :class:`aiocache.plugins.BasePlugin` derived classes.
    :param namespace: string to use as default prefix for the key used in all operations of
        the backend. Default is an empty string, "".
    :param timeout: int or float in seconds specifying maximum timeout for the operations to last.
        By default, its 5.
    :param maxsize: int maximum number of keys to store (None for unlimited)
    """

    NAME = "memory"

    # TODO(PY312): https://peps.python.org/pep-0692/
    def __init__(self, **kwargs):
        # Extract maxsize before passing kwargs to base class
        self.maxsize = kwargs.pop("maxsize", None)
        if "serializer" not in kwargs:
            kwargs["serializer"] = NullSerializer()
        super().__init__(**kwargs)

        self._cache: OrderedDict[str, object] = OrderedDict()
        self._handlers: dict[str, asyncio.TimerHandle] = {}

    def _mark_accessed(self, key: str) -> None:
        """Move key to end to mark as recently used."""
        if key in self._cache:
            self._cache.move_to_end(key)

    def _evict_if_needed(self) -> None:
        """Evict least recently used items if over maxsize."""
        if self.maxsize is None:
            return

        while len(self._cache) > self.maxsize:
            key, _ = self._cache.popitem(last=False)  # Remove LRU item
            if key in self._handlers:
                self._handlers[key].cancel()
                del self._handlers[key]

    async def _get(self, key, encoding="utf-8", _conn=None):
        self._mark_accessed(key)
        return self._cache.get(key)

    async def _gets(self, key, encoding="utf-8", _conn=None):
        return await self._get(key, encoding=encoding, _conn=_conn)

    async def _multi_get(self, keys, encoding="utf-8", _conn=None):
        return [await self._get(key, encoding=encoding, _conn=_conn) for key in keys]

    async def _set(self, key, value, ttl=None, _cas_token=None, _conn=None):
        if _cas_token is not None and self._cache.get(key) != _cas_token:
            return 0

        if key in self._handlers:
            self._handlers[key].cancel()

        self._cache[key] = value
        self._cache.move_to_end(key)

        if ttl:
            loop = asyncio.get_running_loop()
            self._handlers[key] = loop.call_later(ttl, self.__delete, key)

        # Evict the oldest items if over limit
        self._evict_if_needed()
        return True

    async def _multi_set(self, pairs, ttl=None, _conn=None):
        for key, value in pairs:
            await self._set(key, value, ttl=ttl)
        return True

    async def _add(self, key, value, ttl=None, _conn=None):
        if key in self._cache:
            raise ValueError(f"Key {key} already exists, use .set to update")
        return await self._set(key, value, ttl=ttl)

    async def _exists(self, key, _conn=None):
        return key in self._cache

    async def _increment(self, key, delta, _conn=None):
        if key not in self._cache:
            self._cache[key] = delta
        else:
            try:
                self._cache[key] = int(self._cache[key]) + delta
            except ValueError:
                raise TypeError("Value is not an integer") from None
        self._mark_accessed(key)
        return self._cache[key]

    async def _expire(self, key, ttl, _conn=None):
        if key not in self._cache:
            return False

        # Cancel existing timer
        if key in self._handlers:
            self._handlers[key].cancel()

        # Set new timer
        if ttl:
            loop = asyncio.get_running_loop()
            self._handlers[key] = loop.call_later(ttl, self.__delete, key)

        self._mark_accessed(key)
        return True

    async def _delete(self, key, _conn=None):
        return self.__delete(key)

    async def _clear(self, namespace=None, _conn=None):
        if namespace:
            for key in list(self._cache):
                if key.startswith(namespace):
                    self.__delete(key)
        else:
            self._cache = OrderedDict()
            self._handlers = {}
        return True

    async def _raw(self, command, *args, encoding="utf-8", _conn=None, **kwargs):
        return getattr(self._cache, command)(*args, **kwargs)

    async def _redlock_release(self, key, value):
        if self._cache.get(key) == value:
            return self.__delete(key)
        return 0

    def __delete(self, key):
        if self._cache.pop(key, None) is not None:
            handle = self._handlers.pop(key, None)
            if handle:
                handle.cancel()
            return 1

        return 0

    def build_key(self, key: str, namespace: Optional[str] = None) -> str:
        return self._str_build_key(key, namespace)

    @classmethod
    def parse_uri_path(cls, path):
        return {}



import msgspec
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Set,
)


from mkfst.models.http import Headers, Parameters, Query

from .models import Parameter
from .parsed_types import FieldMetadata, FieldsMetadata, FieldType, PropertySchema


def parse_type(field: FieldMetadata) -> List[str] | str:
    field_data: List[FieldType] = field.get("anyOf")
    if field_data:
        return {"enum": [field_type.get("type") for field_type in field_data]}

    return {"type": field.get("type", "string")}


def format_header_name(name: str):
    return "-".join([segment.capitalize() for segment in name.split("_")])


def parse_param(
    param: type[Headers] | type[Parameters] | type[Query],
    path_params: Set[str],
) -> Parameter:
    schema_properties = param.model_json_schema()

    metadata: FieldsMetadata = schema_properties.get("properties")
    required = schema_properties.get("required", [])

    fields = list(param.__struct_fields__)

    if param in [Parameters]:
        fields = [field for field in fields if field in path_params]

        additional_params = list(
            filter(lambda path_param: path_param not in fields, path_params)
        )

        for path_param in additional_params:
            metadata[path_param] = {
                "type": "string",
                "format": "string",
            }

            fields.append(path_param)

    location: str | None = None

    if param in [Headers] or param in Headers.__subclasses__():
        location = "header"

    elif param in [Query] or param in Query.__subclasses__():
        location = "query"

    elif param in [Parameters] or param in Parameters.__subclasses__():
        location = "path"

    else:
        location = "cookie"

    return {
        format_header_name(field): {
            "name": format_header_name(field) if location == "header" else field,
            "description": f"{field} header",
            "required": field in required,
            "in": location,
            "schema": {
                **parse_type(metadata.get(field)),
                "format": metadata.get(field, {}).get("format"),
            },
        }
        for field in fields
    }



import re
from datetime import timedelta


class TimeParser:
    def __init__(self) -> None:
        self._units = {
            "s": "seconds",
            "m": "minutes",
            "h": "hours",
            "d": "days",
            "w": "weeks",
        }

    def parse(self, time_amount: str):
        return float(
            timedelta(
                **{
                    self._units.get(
                        m.group("unit").lower(), 
                        "seconds"
                    ): float(
                        m.group("val")
                    )
                    for m in re.finditer(
                        r"(?P<val>\d+(\.\d+)?)(?P<unit>[smhdw]?)",
                        time_amount,
                        flags=re.I,
                    )
                }
            ).total_seconds()
        )



from typing import List

from pydantic import BaseModel, StrictBool, StrictStr


class ParsedOperationMetadata(BaseModel):
    group_name: StrictStr
    name: StrictStr
    tags: List[StrictStr] | None = None
    description: StrictStr | None = None
    summary: StrictStr | None = None
    docs_url: StrictStr | None = None
    docs_description: StrictStr | None = None
    deprecated: StrictBool | None = None




from __future__ import annotations
import msgspec
import threading
import datetime
from typing import Generic, TypeVar
from .entry import Entry
from .log_level import LogLevel


T = TypeVar("T")


class Log(msgspec.Struct, Generic[T], kw_only=True):
    level: LogLevel
    entry: Entry
    filename: str
    function_name: str
    line_number: int
    thread_id: int = msgspec.field(
        default_factory=threading.get_native_id,
    )
    timestamp: str = msgspec.field(
        default_factory=lambda: datetime.datetime.now(datetime.UTC).isoformat()
    )



from __future__ import annotations
import os
import orjson
import msgspec
from ipaddress import IPv4Address, IPv6Address
from typing import Callable, Dict, Literal, Union

PrimaryType = Union[str, int, float, bytes, bool]


class Env(msgspec.Struct, kw_only=True):
    MERCURY_SYNC_SERVER_URL: str | None = None
    MERCURY_SYNC_API_VERISON: str = "0.0.1"
    MERCURY_SYNC_TASK_EXECUTOR_TYPE: Literal["thread", "process", "none"] = "process"
    MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_REJECTION_SENSITIVITY: float = 2
    MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_WINDOW: str = "1m"
    MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_THRESHOLD: Union[int, float] = 0.2
    MERCURY_SYNC_HTTP_HANDLER_TIMEOUT: str = "1m"
    MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY: Literal[
        "none", "global", "endpoint", "ip", "ip-endpoint", "custom"
    ] = "none"
    MERCURY_SYNC_HTTP_RATE_LIMITER_TYPE: Literal[
        "adaptive",
        "cpu-adaptive",
        "leaky-bucket",
        "rate-adaptive",
        "sliding-window",
        "token-bucket",
    ] = "sliding-window"
    MERCURY_SYNC_HTTP_CORS_ENABLED: bool = False
    MERCURY_SYNC_HTTP_MEMORY_LIMIT: str = "512mb"
    MERCURY_SYNC_HTTP_CPU_LIMIT: Union[float, int] = 50
    MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF_RATE: int = 10
    MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF: str = "1s"
    MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD: str = "1s"
    MERCURY_SYNC_HTTP_RATE_LIMIT_REQUESTS: int = 100
    MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY: Literal[
        "ip",
        "endpoint",
        "global",
        "ip-endpoint",
        "none",
    ] = "none"
    MERCURY_SYNC_HTTP_RATE_LIMIT_DEFAULT_REJECT: bool = True
    MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION: bool = False
    MERCURY_SYNC_USE_HTTP_SERVER: bool = True
    MERCURY_SYNC_USE_HTTP_AND_TCP_SERVERS: bool = False
    MERCURY_SYNC_USE_UDP_MULTICAST: bool = False
    MERCURY_SYNC_TCP_CONNECT_RETRIES: int = 3
    MERCURY_SYNC_CLEANUP_INTERVAL: str = "0.25s"
    MERCURY_SYNC_MAX_CONCURRENCY: int = 2048
    MERCURY_SYNC_AUTH_SECRET: str = "testtoken"
    MERCURY_SYNC_MULTICAST_GROUP: IPv4Address | IPv6Address = "224.1.1.1"
    MERCURY_SYNC_LOGS_DIRECTORY: str = os.getcwd()
    MERCURY_SYNC_REQUEST_TIMEOUT: str = "30s"
    MERCURY_SYNC_LOG_LEVEL: str = "info"
    MERCURY_SYNC_TASK_RUNNER_MAX_THREADS: int = os.cpu_count()
    MERCURY_SYNC_MAX_REQUEST_CACHE_SIZE: int = 100
    MERCURY_SYNC_ENABLE_REQUEST_CACHING: bool = False
    MERCURY_SYNC_VERIFY_SSL_CERT: Literal["REQUIRED", "OPTIONAL", "NONE"] = "REQUIRED"

    @classmethod
    def types_map(cls) -> Dict[str, Callable[[str], PrimaryType]]:
        return {
            "MERCURY_SYNC_SERVER_URL": str,
            "MERCURY_SYNC_API_VERISON": str,
            "MERCURY_SYNC_TASK_EXECUTOR_TYPE": str,
            "MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_REJECTION_SENSITIVITY": float,
            "MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_WINDOW": str,
            "MERCURY_SYNC_HTTP_HANDLER_TIMEOUT": str,
            "MERCURY_SYNC_USE_UDP_MULTICAST": lambda value: True
            if value.lower() == "true"
            else False,
            "MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_THRESHOLD": float,
            "MERCURY_SYNC_HTTP_CORS_ENABLED": lambda value: True
            if value.lower() == "true"
            else False,
            "MERCURY_SYNC_HTTP_MEMORY_LIMIT": str,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF_RATE": int,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF": str,
            "MERCURY_SYNC_HTTP_CPU_LIMIT": float,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY": str,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD": str,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_REQUESTS": int,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_DEFAULT_REJECT": lambda value: True
            if value.lower() == "true"
            else False,
            "MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION": lambda value: True
            if value.lower() == "true"
            else False,
            "MERCURY_SYNC_USE_HTTP_SERVER": lambda value: True
            if value.lower() == "true"
            else False,
            "MERCURY_SYNC_TCP_CONNECT_RETRIES": int,
            "MERCURY_SYNC_CLEANUP_INTERVAL": str,
            "MERCURY_SYNC_MAX_CONCURRENCY": int,
            "MERCURY_SYNC_AUTH_SECRET": str,
            "MERCURY_SYNC_MULTICAST_GROUP": str,
            "MERCURY_SYNC_LOGS_DIRECTORY": str,
            "MERCURY_SYNC_REQUEST_TIMEOUT": str,
            "MERCURY_SYNC_LOG_LEVEL": str,
            "MERCURY_SYNC_TASK_RUNNER_MAX_THREADS": int,
            "MERCURY_SYNC_MAX_REQUEST_CACHE_SIZE": int,
            "MERCURY_SYNC_ENABLE_REQUEST_CACHING": str,
        }

    def model_dump(self, exclude_none: bool = False):
        if exclude_none:
            return {
                key: value
                for key, value in msgspec.structs.asdict(self)
                if value is not None
            }

        return msgspec.structs.asdict(self)

    def model_dump_json(self, exclude_none: bool = False):
        if exclude_none:
            return orjson.dumps(
                {
                    key: value
                    for key, value in msgspec.structs.asdict(self)
                    if value is not None
                }
            )

        return orjson.dumps(msgspec.structs.asdict(self))



from typing import Dict

from mkfst.logging.models import LogLevel


class LogLevelMap:
    def __init__(self) -> None:
        self._levels: Dict[LogLevel, int] = {
            LogLevel.TRACE: 0,
            LogLevel.DEBUG: 1,
            LogLevel.INFO: 2,
            LogLevel.WARN: 3,
            LogLevel.ERROR: 4,
            LogLevel.CRITICAL: 5,
            LogLevel.FATAL: 6,
        }

    def __getitem__(self, level: LogLevel):
        return self._levels[level]



import datetime
import glob
import os
import pathlib
import re
from typing import Dict, Literal

from mkfst.logging.rotation import (
    FileSizeParser,
    TimeParser,
)

RetentionPolicyConfig = Dict[Literal["max_age", "rotation_time", "max_size"], str]

ParsedRetentionPolicyConfig = Dict[
    Literal["max_age", "rotation_time", "max_size"], int | float | str
]

CheckState = Dict[Literal["max_age", "rotation_time", "max_size"], bool]

PolicyQuery = Dict[
    Literal["file_age", "file_size", "logfile_path"], int | float | pathlib.Path
]


def get_timestamp(filenamne: str):
    if match := re.match(r"[+-]?([0-9]*[.])?[0-9]+", filenamne):
        return float(match.group(0))

    return 0


class RetentionPolicy:
    def __init__(self, retention_policy: RetentionPolicyConfig) -> None:
        self._retention_policy = retention_policy
        self._parsed_policy: ParsedRetentionPolicyConfig = {}

        self._time_parser = TimeParser()
        self._file_size_parser = FileSizeParser()

    def parse(self):
        if max_age := self._retention_policy.get("max_age"):
            self._parsed_policy["max_age"] = self._time_parser.parse(max_age)

        if max_size := self._retention_policy.get("max_size"):
            self._parsed_policy["max_size"] = self._file_size_parser.parse(max_size)

        if rotation_time := self._retention_policy.get("rotation_time"):
            self._parsed_policy["rotation_time"] = rotation_time

    def matches_policy(
        self,
        policy_query: PolicyQuery,
    ):
        checks: CheckState = {
            "file_age": True,
            "file_size": True,
            "rotation_time": True,
        }

        resolved_path: pathlib.Path = policy_query["logfile_path"]
        logfile_directory = str(resolved_path.parent.absolute().resolve())

        if (file_age := policy_query["file_age"]) and (
            max_age := self._parsed_policy.get("max_age")
        ):
            checks["file_age"] = file_age < max_age

        if (file_size := policy_query["file_size"]) and (
            max_file_size := self._parsed_policy.get("max_size")
        ):
            checks["file_size"] = file_size < max_file_size

        if rotation_time := self._parsed_policy.get("rotation_time"):
            last_archived: datetime.datetime | None = None
            existing_logfiles = glob.glob(
                os.path.join(logfile_directory, f"{resolved_path.stem}_*_archived.zst")
            )

            if len(existing_logfiles) > 0:
                last_archived = datetime.datetime.fromtimestamp(
                    os.path.getmtime(
                        max(
                            existing_logfiles,
                            key=os.path.getmtime,
                        )
                    ),
                )

            current_time = datetime.datetime.now()
            current_time_string = current_time.strftime("%H:%M")

            if last_archived and (current_time - last_archived).seconds < 60:
                checks["rotation_time"] = True

            else:
                checks["rotation_time"] = rotation_time != current_time_string

        return len([check for check in checks.values() if check is True]) >= len(
            self._parsed_policy
        )



from typing import (
    Any,
    List,
    Literal,
    Optional,
    Union,
)

from mkfst.logging import Logger, LogLevel
from mkfst.middleware.base import Middleware
from mkfst.middleware.base.response_context import ResponseContext
from mkfst.middleware.base.types import Handler, MiddlewareHandler, MiddlewareResult
from mkfst.models.logging import Event

from .cors_headers import CorsHeaders


class Cors(Middleware):
    def __init__(
        self,
        access_control_allow_origin: List[str] = None,
        access_control_allow_methods: List[
            Literal["GET", "HEAD", "OPTIONS", "POST", "PUT", "PATCH", "DELETE", "TRACE"]
        ] = None,
        access_control_expose_headers: Optional[List[str]] = None,
        access_control_max_age: Optional[Union[int, float]] = None,
        access_control_allow_credentials: Optional[bool] = None,
        access_control_allow_headers: Optional[List[str]] = None,
    ) -> None:
        self._cors_config = CorsHeaders(
            access_control_allow_origin=access_control_allow_origin,
            access_control_expose_headers=access_control_expose_headers,
            access_control_max_age=access_control_max_age,
            access_control_allow_credentials=access_control_allow_credentials,
            access_control_allow_methods=access_control_allow_methods,
            access_control_allow_headers=access_control_allow_headers,
        )

        self.origins = self._cors_config.access_control_allow_origin
        self.cors_methods = self._cors_config.access_control_allow_methods
        self.cors_headers = self._cors_config.access_control_allow_headers
        self.allow_credentials = self._cors_config.access_control_allow_credentials

        self.allow_all_origins = "*" in self._cors_config.access_control_allow_origin

        allowed_headers = self._cors_config.access_control_allow_headers
        self.allow_all_headers = False

        if allowed_headers:
            self.allow_all_headers = "*" in allowed_headers

        self.simple_headers = self._cors_config.to_simple_headers()
        self.preflight_headers = self._cors_config.to_preflight_headers()
        self.preflight_explicit_allow_origin = (
            not self.allow_all_origins or self.allow_credentials
        )

        self._logger = Logger()

        super().__init__(
            self.__class__.__name__,
            methods=["OPTIONS"],
            response_headers=self._cors_config.to_headers(),
        )

    async def __run__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ) -> MiddlewareResult:
        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            
            await ctx.log(Event(
                level=LogLevel.DEBUG,
                message=f'Request - {context.method} {context.path}:{context.ip_address} - Verifying request meets CORS policy'
            ))
            
            headers = context.get_headers()
            method = context.method

            if headers:            
                parsed_headers = headers.model_dump()

            else:
                parsed_headers = {}

            origin = parsed_headers.get("origin")
            access_control_request_method: str | None = parsed_headers.get("access-control-request-method")
            access_control_request_headers: str | None = parsed_headers.get("access-control-request-headers", "")

            if method == "OPTIONS" and access_control_request_method:
                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Verifying OPTIONS request meets CORS policy'
                ))

                response_headers = dict(self.preflight_headers)

                failures: List[str] = []

                if self.allow_all_origins is False and origin not in self.origins:
                    failures.append("origin")
                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Origin of {origin} failed Allowed Origin policy'
                    ))

                elif self.preflight_explicit_allow_origin:
                    response["Access-Control-Allow-Origin"] = origin

                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Adding preflight origin of {origin}'
                    ))

                if access_control_request_method not in self.cors_methods:
                    failures.append("method")
                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Method of {method} failed Allowed Methods policy'
                    ))

                if self.allow_all_headers and access_control_request_headers is not None:
                    response_headers["Access-Control-Allow-Headers"] = (
                        access_control_request_headers
                    )
                    await ctx.log(Event(
                        level=LogLevel.DEBUG,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Adding CORS headers to response'
                    ))

                elif access_control_request_headers:
                    for header in access_control_request_headers.split(","):
                        if header.lower().strip() not in self.cors_headers:
                            failures.append("headers")
                            await ctx.log(Event(
                                level=LogLevel.DEBUG,
                                message=f'Request - {context.method} {context.path}:{context.ip_address} - Header of {header} failed Allowed Headers policy'
                            ))

                            break

                if len(failures) > 0:
                    failures_message = ", ".join(failures)
                    context.status = 401

                    context.errors.append(
                        Exception(failures_message)
                    )

                    await ctx.log(Event(
                        level=LogLevel.ERROR,
                        message=f'Request - {context.method} {context.path}:{context.ip_address} - Rejected by CORS policy with status of {context.status}'
                    ))

                    return (
                        context,
                        f"Disallowed CORS {failures_message}",
                    ), False


                context.response_headers.update(response_headers)

                context.errors.append(
                    Exception('Rejected by CORS policy')
                )

                await ctx.log(Event(
                    level=LogLevel.ERROR,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Rejected by CORS policy with status of {context.status}'
                ))

                return (
                    context,
                    response,
                ), False

            response_headers = dict(self.simple_headers)

            if self.allow_all_origins and parsed_headers.get("cookie"):
                response_headers["access-control-allow-origin"] = origin

                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Adding allowed origin of {origin}'
                ))

            elif origin in self.origins:
                response_headers["access-control-allow-origin"] = origin

                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {context.method} {context.path}:{context.ip_address} - Adding allowed origin of {origin}'
                ))

            context.response_headers.update(response_headers)

            await ctx.log(Event(
                level=LogLevel.DEBUG,
                message=f'Request - {context.method} {context.path}:{context.ip_address} - Request met CORS policy'
            ))

            return (
                context,
                response,
            ), True



from __future__ import annotations
import re
from http.cookies import SimpleCookie
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Tuple,
    Type,
    get_origin,
    get_args,
)

import orjson

from mkfst.models import (
    HTML,
    Body,
    Cookies,
    FileUpload,
    Headers,
    Parameters,
    Query,
    Model,
)


KeyType = Literal["positional", "keyword"]
Data = Model | Body | HTML | FileUpload | str | bytes | list | dict
Json = dict | list
Raw = str | bytes


class Fabricator:
    __slots__ = (
        "optional",
        "required",
        "param_keys",
        "_params",
        "_http_pattern",
        "_args_count",
        "_body_type",
        "_ordered_params",
        "_unordered_params",
        "_headers_key",
        "_headers_key_type",
        "_params_key",
        "_params_key_type",
        "_query_key",
        "_query_key_type",
        "_cookies_key",
        "_cookie_key_type",
        "_body_key",
        "_body_key_type",
        "headers",
        "params",
        "query",
        "cookies",
        "body",
    )

    def __init__(
        self,
        required_params: List[
            Tuple[
                str,
                Headers | Parameters | Query | Model | HTML | FileUpload | Cookies,
            ]
        ],
        optional_params: Dict[
            str,
            Headers | Parameters | Query | Model | HTML | FileUpload | Cookies,
        ],
    ) -> None:
        self.optional = {
            key: value[0] for key, value in optional_params.items() if len(value) > 0
        }

        self.required = required_params

        self._http_pattern = re.compile(r"https:\\\\|http:\\\\")
        self._args_count = len(required_params) + len(optional_params)

        self._headers_key: int | str | None = None
        self._headers_key_type: KeyType | None = None

        self._params_key: int | str | None = None
        self._params_key_type: KeyType | None = None

        self._query_key: int | str | None = None
        self._query_key_type: KeyType | None = None

        self._cookies_key: int | str | None = None
        self._cookie_key_type: KeyType | None = None

        self._body_key: int | str | None = None
        self._body_key_type: KeyType | None = None

        self._body_type: Literal[
            "file",
            "html",
            "model",
            "json",
            "body",
            "raw",
        ] = "model"

        self._params: Dict[
            Literal[
                "headers",
                "parameters",
                "query",
                "body",
                "cookies",
            ],
            Tuple[
                Headers | Parameters | Query | Model | HTML | FileUpload | Cookies,
                int | None,
            ],
        ] = {}

        self._ordered_params: List[
            Tuple[
                Literal[
                    "headers",
                    "parameters",
                    "query",
                    "body",
                    "cookies",
                ],
                Headers
                | Parameters
                | Query
                | Model
                | Body
                | HTML
                | FileUpload
                | Cookies,
            ]
        ] = []

        self._unordered_params: List[
            Tuple[
                Literal[
                    "headers",
                    "parameters",
                    "query",
                    "body",
                    "cookies",
                ],
                Headers
                | Parameters
                | Query
                | Model
                | Body
                | HTML
                | FileUpload
                | Cookies,
                str,
            ]
        ] = []

        self.param_keys: Dict[
            Literal[
                "headers",
                "parameters",
                "query",
                "body",
                "cookies",
            ],
            int | str,
        ] = {}

        for position in range(len(self.required)):
            _, annotation = self.required[position]
            self._parse_to_param(annotation, position=position)

        for name, annotation in self.optional.items():
            self._parse_to_param(annotation, name=name)

        self.headers: Headers | None = None
        self.params: Parameters | None = None
        self.query: Query | None = None
        self.cookies: Cookies | None = None
        self.body: Data | None = None

    def __getitem__(
        self,
        param_name: Literal[
            "headers",
            "parameters",
            "query",
            "body",
            "cookies",
        ],
    ):
        if param := self._params.get(param_name):
            return param[0]

    @property
    def required_params(self):
        param_type_map = {
            Parameters: "path",
            Query: "query",
            Headers: "headers",
        }

        return [param_type_map.get(value, "body") for _, value in self.required]

    def _parse_to_param(
        self,
        annotation: Type[Headers]
        | Type[Parameters]
        | Type[Query]
        | Type[Data]
        | Type[Body]
        | Type[HTML]
        | Type[FileUpload]
        | Type[Cookies],
        position: int | None = None,
        name: str | None = None,
    ):
        param_type: str | None = None

        if annotation in Headers.__subclasses__():
            self._params["headers"] = (annotation, position)

            self._headers_key = position if position is not None else name
            self._headers_key_type = "positional" if position else "keyword"

            param_type = "headers"

        elif annotation in Cookies.__subclasses__():
            self._params["cookies"] = (annotation, position)

            self._cookies_key = position if position is not None else name
            self._cookie_key_type = "positional" if position is not None else "keyword"

            param_type = "cookies"

        elif annotation in Parameters.__subclasses__():
            self._params["parameters"] = (annotation, position)

            self._params_key = position if position is not None else name
            self._params_key_type = "positional" if position is not None else "keyword"

            param_type = "parameters"

        elif annotation in Query.__subclasses__():
            self._params["query"] = (annotation, position)

            self._query_key = position if position is not None else name
            self._query_key_type = "positional" if position is not None else "keyword"

            param_type = "query"

        elif annotation == FileUpload or annotation in FileUpload.__subclasses__():
            self._params["body"] = (
                FileUpload,
                position,
            )

            self._body_key = position if position is not None else name
            self._body_key_type = "positional" if position is not None else "keyword"

            param_type = "body"
            self._body_type = "file"

        elif annotation == HTML or annotation in HTML.__subclasses__():
            self._params["body"] = (
                HTML,
                position,
            )

            self._body_key = position if position is not None else name
            self._body_key_type = "positional" if position is not None else "keyword"

            param_type = "body"
            self._body_type = "html"

        elif annotation in Model.__subclasses__() and annotation != Body:
            self._params["body"] = (
                annotation,
                position,
            )

            self._body_key = position if position is not None else name
            self._body_key_type = "positional" if position is not None else "keyword"

            param_type = "body"
            self._body_type = "model"

        elif annotation == Body or annotation in Body.__subclasses__():
            self._params["body"] = (
                annotation,
                position,
            )

            self._body_key = position if position is not None else name
            self._body_key_type = "positional" if position is not None else "keyword"
            param_type = "body"
            self._body_type = "body"

        elif annotation in get_args(Json) or get_origin(annotation) in get_args(Json):
            self._params["body"] = (
                annotation,
                position,
            )

            self._body_key = position if position is not None else name
            self._body_key_type = "positional" if position is not None else "keyword"

            param_type = "body"
            self._body_type = "json"

        else:
            self._params["body"] = (
                annotation,
                position,
            )

            self._body_key = position if position is not None else name
            self._body_key_type = "positional" if position is not None else "keyword"

            param_type = "body"
            self._body_type = "raw"

        if position is not None:
            self._ordered_params.append((param_type, annotation))

            self.param_keys[param_type] = position

        else:
            self._unordered_params.append(
                (
                    param_type,
                    annotation,
                    name,
                )
            )

            self.param_keys[param_type] = name

    def parse(
        self,
        request_data: bytes,
        request_headers: dict[str, str] | None = None,
        request_query: str | None = None,
        request_params: Dict[str, str | Tuple[str]] | None = None,
        has_middleware: bool = False,
    ):
        positional_args: List[
            Headers | Parameters | Query | Data | Body | HTML | FileUpload | Cookies,
        ] = []

        keyword_args: Dict[
            str,
            Headers | Parameters | Query | Data | Body | HTML | FileUpload | Cookies,
        ] = {}

        if self._args_count == 0:
            return (
                positional_args,
                keyword_args,
                None,
            )

        headers: Headers | None = None
        parameters: Parameters | None = None
        queries: Query | None = None
        cookies: Cookies | None = None
        body: Any | None = None

        if self._headers_key_type and request_headers:
            (headers, validation_error, positional) = self._parse_headers(
                request_headers,
            )

            if validation_error:
                return (
                    None,
                    None,
                    validation_error,
                )

            if positional:
                positional_args.insert(self._headers_key, headers)

            else:
                keyword_args[self._headers_key] = headers

            self.headers = headers

        if self._body_key_type:
            (body, validation_error, positional) = self._parse_body(
                request_data,
                request_headers,
                has_middleware=has_middleware,
            )

            if validation_error:
                return (
                    None,
                    None,
                    validation_error,
                )

            elif positional:
                positional_args.insert(self._body_key, body)

            else:
                keyword_args[self._body_key] = body

            self.body = body

        if self._cookie_key_type and cookies:
            (
                cookies,
                validation_error,
                positional,
            ) = self._parse_cookies(
                request_headers,
            )

            if positional and validation_error:
                return (
                    None,
                    None,
                    validation_error,
                )

            elif positional:
                positional_args.insert(self._cookies_key, cookies)

            else:
                keyword_args[self._cookies_key] = cookies

            self.cookies = cookies

        if self._params_key_type:
            (
                parameters,
                validation_error,
                positional,
            ) = self._parse_params(request_params)

            if validation_error:
                return (
                    None,
                    None,
                    validation_error,
                )

            elif positional:
                positional_args.insert(self._params_key, parameters)

            else:
                keyword_args[self._params_key] = parameters

            self.params = parameters

        if self._query_key_type:
            (queries, validation_error, positional) = self._parse_query(request_query)

            if validation_error:
                return (
                    None,
                    None,
                    validation_error,
                )

            elif positional:
                positional_args.insert(self._query_key, queries)

            else:
                keyword_args[self._query_key] = queries

            self.query = queries

        return (positional_args, keyword_args, None)

    def _parse_headers(
        self,
        data: dict[str, str],
    ):
        annotation = self._get_annotation(self._headers_key, self._headers_key_type)

        try:
            return (
                annotation(**data),
                None,
                self._headers_key_type == "positional",
            )

        except Exception as validation_error:
            return (
                data,
                validation_error,
                self._headers_key_type == "positional",
            )

    def _parse_body(
        self,
        request_data: bytes,
        request_headers: Dict[str, str],
        has_middleware: bool = False,
    ):
        annotation = self._get_annotation(self._body_key, self._body_key_type)
        positional = self._body_key_type == "positional"

        if has_middleware:
            return (
                request_data,
                None,
                positional,
            )

        try:
            content_length = int(
                request_headers.get("content-length")
                or request_headers.get("Content-Length")
                or 0
            )
            if len(request_data) < content_length:
                raise Exception("No data")

            match self._body_type:
                case "file":
                    encoding = request_headers.get("content-encoding")

                    body = annotation(
                        data=request_data.strip().decode(encoding=encoding),
                        content_type=request_headers.get("content-type"),
                        content_encoding=encoding,
                    )

                case "html":
                    body = annotation(content=request_data.strip().decode())

                case "model":
                    body = annotation(**orjson.loads(request_data))

                case "json":
                    body = orjson.loads(request_data)

                case "body":
                    body = Body(content=request_data.strip())

                case _:
                    body = request_data.strip()

            return (
                body,
                None,
                positional,
            )

        except Exception as validation_error:
            return (
                None,
                validation_error,
                positional,
            )

    def _parse_cookies(
        self,
        request_headers: Dict[str, str],
    ):
        positional = self._cookie_key_type == "positional"
        try:
            annotation = self._get_annotation(self._cookies_key, self._cookie_key_type)

            cookie_header = request_headers.get("cookie")

            if cookie_header is None:
                return annotation()

            parsed_cookies = SimpleCookie()
            parsed_cookies.load(cookie_header)

            return (
                parsed_cookies,
                None,
                positional,
            )

        except Exception as validation_error:
            return (
                None,
                validation_error,
                positional,
            )

    def _parse_params(
        self,
        params: Dict[str, Any],
    ):
        positional = self._params_key_type == "positional"
        try:
            annotation: Type[Parameters] = self._get_annotation(
                self._params_key, self._params_key_type
            )

            parameters = annotation(**params)

            return (parameters, None, positional)

        except Exception as validation_error:
            return (None, validation_error, positional)

    def _parse_query(self, query: str):
        try:
            annotation = self._get_annotation(self._query_key, self._query_key_type)
            parsed_query = Query.make(annotation, query)

            return (parsed_query, None, self._body_key_type == "positional")

        except Exception as validation_error:
            return (None, validation_error, self._body_key_type == "positional")

    def _get_annotation(
        self,
        key: int | str,
        key_type: KeyType,
    ):
        annotation_data = (
            self._ordered_params[key]
            if key_type == "positional"
            else self._unordered_params[key]
        )

        return annotation_data[1]



import re
from typing import (
    Any,
    Literal,
)
from urllib.parse import urlunparse
from mkfst.logging import Logger, LogLevel
from mkfst.middleware.base import Middleware
from mkfst.middleware.base.response_context import ResponseContext
from mkfst.middleware.base.types import Handler, MiddlewareHandler
from mkfst.models.logging import Event


class HTTPSRedirect(Middleware):
    def __init__(self):
        self._logger = Logger()
        super().__init__(
            self.__class__.__name__,
            methods=[
                "DELETE",
                "GET",
                "HEAD",
                "OPTIONS",
                "PATCH",
                "POST",
                "PUT",
                "TRACE",
            ],
        )

        self._scheme: dict[Literal["http", "ws"], Literal["https", "wss"]] = {
            "http": "https",
            "ws": "wss",
        }

        self._port_pattern = re.compile(r":\d{4}")

    async def __run__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ):
        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            await ctx.log(
                Event(
                    level=LogLevel.DEBUG,
                    message=f"Request - {context.method} {context.path}:{context.ip_address} - Checking for HTTPS redirect",
                )
            )

            try:
                if (
                    context.protocol in ["http", "ws"]
                    and context.upgrade_port is not None
                ):
                    redirect_scheme = self._scheme[context.protocol]

                    url = context.to_request_url()

                    redirect_url = urlunparse(
                        url._replace(
                            scheme=redirect_scheme,
                            netloc=self._port_pattern.sub(
                                f":{context.upgrade_port}",
                                url.netloc,
                                count=1,
                            ),
                        )
                    )

                    context.status = 307
                    context.response_headers.update(
                        {
                            "location": redirect_url,
                        }
                    )

                    await ctx.log(
                        Event(
                            level=LogLevel.INFO,
                            message=f"Request - {context.method} {context.path}:{context.ip_address} - Enforcing {redirect_url} redirect",
                        )
                    )

                    return (
                        context,
                        response,
                    ), False

                else:
                    return (
                        context,
                        response,
                    ), True

            except Exception as e:
                context.errors.append(e)
                context.status = 500

                await ctx.log(
                    Event(
                        level=LogLevel.ERROR,
                        message=f"Request - {context.method} {context.path}:{context.ip_address} - Encountered error attempting HTTPS redirect - {str(e)}",
                    )
                )

                return (
                    context,
                    response,
                ), False

    async def close(self):
        pass

    def abort(self):
        pass



import asyncio
from typing import TypeVar, Generic


T = TypeVar("T")


class ServerState(Generic[T]):
    """
    Shared servers state that is available between all protocol instances.
    """

    def __init__(self) -> None:
        self.total_requests = 0
        self.connections: set[T] = set()
        self.tasks: set[asyncio.Task[None]] = set()
        self.default_headers: list[tuple[bytes, bytes]] = []